# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Conftest module for Pytest."""
import asyncio
import inspect
import logging
import os
import platform
import socket
import sys
import threading
import time
from functools import WRAPPER_ASSIGNMENTS, wraps
from pathlib import Path
from threading import Timer
from types import FunctionType, MethodType
from typing import Callable, List, Optional, Sequence, cast
from unittest.mock import patch

import docker as docker
import gym
import pytest
from docker.models.containers import Container
from oef.agents import AsyncioCore, OEFAgent

from aea import AEA_DIR
from aea.aea import AEA
from aea.cli.utils.config import _init_cli_config
from aea.common import Address
from aea.configurations.base import ComponentType, ConnectionConfig, ContractConfig
from aea.configurations.base import DEFAULT_AEA_CONFIG_FILE as AGENT_YAML
from aea.configurations.base import DEFAULT_CONNECTION_CONFIG_FILE as CONNECTION_YAML
from aea.configurations.base import DEFAULT_CONTRACT_CONFIG_FILE as CONTRACT_YAML
from aea.configurations.base import DEFAULT_PROTOCOL_CONFIG_FILE as PROTOCOL_YAML
from aea.configurations.base import DEFAULT_SKILL_CONFIG_FILE as SKILL_YAML
from aea.configurations.base import PublicId
from aea.configurations.constants import DEFAULT_CONNECTION, DEFAULT_LEDGER
from aea.configurations.loader import load_component_configuration
from aea.connections.base import Connection
from aea.connections.stub.connection import StubConnection
from aea.contracts.base import Contract, contract_registry
from aea.crypto.cosmos import DEFAULT_ADDRESS as COSMOS_DEFAULT_ADDRESS
from aea.crypto.cosmos import _COSMOS
from aea.crypto.ethereum import DEFAULT_ADDRESS as ETHEREUM_DEFAULT_ADDRESS
from aea.crypto.ethereum import _ETHEREUM
from aea.crypto.fetchai import DEFAULT_ADDRESS as FETCHAI_DEFAULT_ADDRESS
from aea.crypto.fetchai import _FETCHAI
from aea.crypto.helpers import PRIVATE_KEY_PATH_SCHEMA
from aea.crypto.registries import make_crypto
from aea.crypto.wallet import CryptoStore
from aea.identity.base import Identity
from aea.test_tools.click_testing import CliRunner as ImportedCliRunner
from aea.test_tools.constants import DEFAULT_AUTHOR

from packages.fetchai.connections.local.connection import LocalNode, OEFLocalConnection
from packages.fetchai.connections.oef.connection import OEFConnection
from packages.fetchai.connections.p2p_libp2p.connection import (
    MultiAddr,
    P2PLibp2pConnection,
)
from packages.fetchai.connections.p2p_libp2p_client.connection import (
    P2PLibp2pClientConnection,
)
from packages.fetchai.connections.tcp.tcp_client import TCPClientConnection
from packages.fetchai.connections.tcp.tcp_server import TCPServerConnection

from tests.data.dummy_connection.connection import DummyConnection  # type: ignore


logger = logging.getLogger(__name__)

CliRunner = ImportedCliRunner

CUR_PATH = os.path.dirname(inspect.getfile(inspect.currentframe()))  # type: ignore
ROOT_DIR = os.path.join(CUR_PATH, "..")
CLI_LOG_OPTION = ["-v", "OFF"]

AUTHOR = DEFAULT_AUTHOR
CONFIGURATION_SCHEMA_DIR = os.path.join(AEA_DIR, "configurations", "schemas")
AGENT_CONFIGURATION_SCHEMA = os.path.join(
    CONFIGURATION_SCHEMA_DIR, "aea-config_schema.json"
)
SKILL_CONFIGURATION_SCHEMA = os.path.join(
    CONFIGURATION_SCHEMA_DIR, "skill-config_schema.json"
)
CONNECTION_CONFIGURATION_SCHEMA = os.path.join(
    CONFIGURATION_SCHEMA_DIR, "connection-config_schema.json"
)
PROTOCOL_CONFIGURATION_SCHEMA = os.path.join(
    CONFIGURATION_SCHEMA_DIR, "protocol-config_schema.json"
)
CONTRACT_CONFIGURATION_SCHEMA = os.path.join(
    CONFIGURATION_SCHEMA_DIR, "contract-config_schema.json"
)
PROTOCOL_SPEC_CONFIGURATION_SCHEMA = os.path.join(
    CONFIGURATION_SCHEMA_DIR, "protocol-specification_schema.json"
)

DUMMY_ENV = gym.GoalEnv

# Ledger identifiers
COSMOS = _COSMOS
ETHEREUM = _ETHEREUM
FETCHAI = _FETCHAI

COSMOS_PRIVATE_KEY_FILE_CONNECTION = "cosmos_connection_private_key.txt"
FETCHAI_PRIVATE_KEY_FILE_CONNECTION = "fetchai_connection_private_key.txt"

COSMOS_PRIVATE_KEY_FILE = PRIVATE_KEY_PATH_SCHEMA.format(COSMOS)
ETHEREUM_PRIVATE_KEY_FILE = PRIVATE_KEY_PATH_SCHEMA.format(ETHEREUM)
FETCHAI_PRIVATE_KEY_FILE = PRIVATE_KEY_PATH_SCHEMA.format(FETCHAI)

# private keys with value on testnet
COSMOS_PRIVATE_KEY_PATH = os.path.join(
    ROOT_DIR, "tests", "data", COSMOS_PRIVATE_KEY_FILE
)
ETHEREUM_PRIVATE_KEY_PATH = os.path.join(
    ROOT_DIR, "tests", "data", ETHEREUM_PRIVATE_KEY_FILE
)
FETCHAI_PRIVATE_KEY_PATH = os.path.join(
    ROOT_DIR, "tests", "data", FETCHAI_PRIVATE_KEY_FILE
)
DEFAULT_PRIVATE_KEY_PATH = COSMOS_PRIVATE_KEY_PATH
FUNDED_ETH_PRIVATE_KEY_1 = (
    "0xa337a9149b4e1eafd6c21c421254cf7f98130233595db25f0f6f0a545fb08883"
)
FUNDED_ETH_PRIVATE_KEY_2 = (
    "0x04b4cecf78288f2ab09d1b4c60219556928f86220f0fb2dcfc05e6a1c1149dbf"
)
FUNDED_ETH_PRIVATE_KEY_3 = (
    "0x6F611408F7EF304947621C51A4B7D84A13A2B9786E9F984DA790A096E8260C64"
)
NON_FUNDED_COSMOS_PRIVATE_KEY_1 = (
    "81b0352f99a08a754b56e529dda965c4ce974edb6db7e90035e01ed193e1b7bc"
)
NON_FUNDED_FETCHAI_PRIVATE_KEY_1 = (
    "b6ef49c3078f300efe2d4480e179362bd39f20cbb2087e970c8f345473661aa5"
)

# addresses with no value on testnet
COSMOS_ADDRESS_ONE = "cosmos1z4ftvuae5pe09jy2r7udmk6ftnmx504alwd5qf"
COSMOS_ADDRESS_TWO = "cosmos1gssy8pmjdx8v4reg7lswvfktsaucp0w95nk78m"
ETHEREUM_ADDRESS_ONE = "0x46F415F7BF30f4227F98def9d2B22ff62738fD68"
ETHEREUM_ADDRESS_TWO = "0x7A1236d5195e31f1F573AD618b2b6FEFC85C5Ce6"
FETCHAI_ADDRESS_ONE = "fetch1paqxtqnfh7da7z9c05l3y3lahe8rhd0nm0jk98"
FETCHAI_ADDRESS_TWO = "fetch19j4dc3e6fgle98pj06l5ehhj6zdejcddx7teac"

# P2P addresses
COSMOS_P2P_ADDRESS = "/dns4/127.0.0.1/tcp/9000/p2p/16Uiu2HAmAzvu5uNbcnD2qaqrkSULhJsc6GJUg3iikWerJkoD72pr"  # relates to NON_FUNDED_COSMOS_PRIVATE_KEY_1
FETCHAI_P2P_ADDRESS = "/dns4/127.0.0.1/tcp/9000/p2p/16Uiu2HAmLBCAqHL8SuFosyDhAKYsLKXBZBWXBsB9oFw2qU4Kckun"  # relates to NON_FUNDED_FETCHAI_PRIVATE_KEY_1
NON_GENESIS_CONFIG = {
    "delegate_uri": "127.0.0.1:11001",
    "entry_peers": [COSMOS_P2P_ADDRESS],
    "local_uri": "127.0.0.1:9001",
    "log_file": "libp2p_node.log",
    "public_uri": "127.0.0.1:9001",
    "ledger_id": "cosmos",
}
NON_GENESIS_CONFIG_TWO = {
    "delegate_uri": "127.0.0.1:11002",
    "entry_peers": [COSMOS_P2P_ADDRESS],
    "local_uri": "127.0.0.1:9002",
    "log_file": "libp2p_node.log",
    "public_uri": "127.0.0.1:9002",
    "ledger_id": "cosmos",
}
PUBLIC_DHT_P2P_MADDR_1 = "/dns4/agents-p2p-dht.sandbox.fetch-ai.com/tcp/9000/p2p/16Uiu2HAkw1ypeQYQbRFV5hKUxGRHocwU5ohmVmCnyJNg36tnPFdx"
PUBLIC_DHT_P2P_MADDR_2 = "/dns4/agents-p2p-dht.sandbox.fetch-ai.com/tcp/9001/p2p/16Uiu2HAmVWnopQAqq4pniYLw44VRvYxBUoRHqjz1Hh2SoCyjbyRW"
PUBLIC_DHT_DELEGATE_URI_1 = "agents-p2p-dht.sandbox.fetch-ai.com:11000"
PUBLIC_DHT_DELEGATE_URI_2 = "agents-p2p-dht.sandbox.fetch-ai.com:11001"

# testnets
COSMOS_TESTNET_CONFIG = {"address": COSMOS_DEFAULT_ADDRESS}
ETHEREUM_TESTNET_CONFIG = {
    "address": ETHEREUM_DEFAULT_ADDRESS,
    "gas_price": 50,
}
FETCHAI_TESTNET_CONFIG = {"address": FETCHAI_DEFAULT_ADDRESS}

# common public ids used in the tests
UNKNOWN_PROTOCOL_PUBLIC_ID = PublicId("unknown_author", "unknown_protocol", "0.1.0")
UNKNOWN_CONNECTION_PUBLIC_ID = PublicId("unknown_author", "unknown_connection", "0.1.0")
UNKNOWN_SKILL_PUBLIC_ID = PublicId("unknown_author", "unknown_skill", "0.1.0")
LOCAL_CONNECTION_PUBLIC_ID = PublicId("fetchai", "local", "0.1.0")
P2P_CLIENT_CONNECTION_PUBLIC_ID = PublicId("fetchai", "p2p_client", "0.1.0")
HTTP_CLIENT_CONNECTION_PUBLIC_ID = PublicId.from_str("fetchai/http_client:0.10.0")
HTTP_PROTOCOL_PUBLIC_ID = PublicId("fetchai", "http", "0.1.0")
STUB_CONNECTION_PUBLIC_ID = DEFAULT_CONNECTION
DUMMY_PROTOCOL_PUBLIC_ID = PublicId("dummy_author", "dummy", "0.1.0")
DUMMY_CONNECTION_PUBLIC_ID = PublicId("dummy_author", "dummy", "0.1.0")
DUMMY_SKILL_PUBLIC_ID = PublicId("dummy_author", "dummy", "0.1.0")

DUMMY_SKILL_PATH = os.path.join(CUR_PATH, "data", "dummy_skill", SKILL_YAML)

MAX_FLAKY_RERUNS = 3
MAX_FLAKY_RERUNS_ETH = 1
MAX_FLAKY_RERUNS_INTEGRATION = 1

FETCHAI_PREF = os.path.join(ROOT_DIR, "packages", "fetchai")
PROTOCOL_SPECS_PREF_1 = os.path.join(ROOT_DIR, "examples", "protocol_specification_ex")
PROTOCOL_SPECS_PREF_2 = os.path.join(ROOT_DIR, "tests", "data")

contract_config_files = [
    os.path.join(FETCHAI_PREF, "contracts", "erc1155", CONTRACT_YAML),
    os.path.join(ROOT_DIR, "tests", "data", "dummy_contract", CONTRACT_YAML),
]

protocol_config_files = [
    os.path.join(ROOT_DIR, "aea", "protocols", "default", PROTOCOL_YAML),
    os.path.join(ROOT_DIR, "aea", "protocols", "scaffold", PROTOCOL_YAML),
    os.path.join(ROOT_DIR, "aea", "protocols", "signing", PROTOCOL_YAML),
    os.path.join(ROOT_DIR, "aea", "protocols", "state_update", PROTOCOL_YAML),
    os.path.join(FETCHAI_PREF, "protocols", "contract_api", PROTOCOL_YAML),
    os.path.join(FETCHAI_PREF, "protocols", "fipa", PROTOCOL_YAML),
    os.path.join(FETCHAI_PREF, "protocols", "gym", PROTOCOL_YAML),
    os.path.join(FETCHAI_PREF, "protocols", "http", PROTOCOL_YAML),
    os.path.join(FETCHAI_PREF, "protocols", "ledger_api", PROTOCOL_YAML),
    os.path.join(FETCHAI_PREF, "protocols", "ml_trade", PROTOCOL_YAML),
    os.path.join(FETCHAI_PREF, "protocols", "oef_search", PROTOCOL_YAML),
    os.path.join(FETCHAI_PREF, "protocols", "tac", PROTOCOL_YAML),
]

connection_config_files = [
    os.path.join(ROOT_DIR, "aea", "connections", "scaffold", CONNECTION_YAML),
    os.path.join(ROOT_DIR, "aea", "connections", "stub", CONNECTION_YAML),
    os.path.join(FETCHAI_PREF, "connections", "gym", CONNECTION_YAML),
    os.path.join(FETCHAI_PREF, "connections", "http_client", CONNECTION_YAML),
    os.path.join(FETCHAI_PREF, "connections", "http_server", CONNECTION_YAML),
    os.path.join(FETCHAI_PREF, "connections", "ledger", CONNECTION_YAML),
    os.path.join(FETCHAI_PREF, "connections", "local", CONNECTION_YAML),
    os.path.join(FETCHAI_PREF, "connections", "oef", CONNECTION_YAML),
    os.path.join(FETCHAI_PREF, "connections", "p2p_libp2p", CONNECTION_YAML),
    os.path.join(FETCHAI_PREF, "connections", "p2p_libp2p_client", CONNECTION_YAML),
    os.path.join(FETCHAI_PREF, "connections", "p2p_stub", CONNECTION_YAML),
    os.path.join(FETCHAI_PREF, "connections", "soef", CONNECTION_YAML),
    os.path.join(FETCHAI_PREF, "connections", "tcp", CONNECTION_YAML),
    os.path.join(FETCHAI_PREF, "connections", "webhook", CONNECTION_YAML),
    os.path.join(CUR_PATH, "data", "dummy_connection", CONNECTION_YAML),
    os.path.join(CUR_PATH, "data", "gym-connection.yaml"),
]


skill_config_files = [
    os.path.join(ROOT_DIR, "aea", "skills", "error", SKILL_YAML),
    os.path.join(ROOT_DIR, "aea", "skills", "scaffold", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "aries_alice", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "aries_faber", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "carpark_client", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "carpark_detection", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "echo", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "erc1155_client", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "erc1155_deploy", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "generic_buyer", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "generic_seller", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "gym", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "http_echo", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "ml_data_provider", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "ml_train", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "simple_service_registration", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "tac_control", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "tac_control_contract", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "tac_negotiation", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "tac_participation", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "thermometer", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "thermometer_client", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "weather_client", SKILL_YAML),
    os.path.join(FETCHAI_PREF, "skills", "weather_station", SKILL_YAML),
    DUMMY_SKILL_PATH,
    os.path.join(CUR_PATH, "data", "dummy_aea", "skills", "dummy", SKILL_YAML),
    os.path.join(CUR_PATH, "data", "dependencies_skill", SKILL_YAML),
    os.path.join(CUR_PATH, "data", "exception_skill", SKILL_YAML),
]


agent_config_files = [
    os.path.join(CUR_PATH, "data", "dummy_aea", AGENT_YAML),
    os.path.join(CUR_PATH, "data", "aea-config.example.yaml"),
    os.path.join(CUR_PATH, "data", "aea-config.example_w_keys.yaml"),
    os.path.join(CUR_PATH, "data", "aea-config.example_multipage.yaml"),
    os.path.join(FETCHAI_PREF, "agents", "aries_alice", AGENT_YAML),
    os.path.join(FETCHAI_PREF, "agents", "aries_faber", AGENT_YAML),
    os.path.join(FETCHAI_PREF, "agents", "car_data_buyer", AGENT_YAML),
    os.path.join(FETCHAI_PREF, "agents", "car_detector", AGENT_YAML),
    os.path.join(FETCHAI_PREF, "agents", "erc1155_client", AGENT_YAML),
    os.path.join(FETCHAI_PREF, "agents", "erc1155_deployer", AGENT_YAML),
    os.path.join(FETCHAI_PREF, "agents", "generic_buyer", AGENT_YAML),
    os.path.join(FETCHAI_PREF, "agents", "generic_seller", AGENT_YAML),
    os.path.join(FETCHAI_PREF, "agents", "gym_aea", AGENT_YAML),
    os.path.join(FETCHAI_PREF, "agents", "ml_data_provider", AGENT_YAML),
    os.path.join(FETCHAI_PREF, "agents", "ml_model_trainer", AGENT_YAML),
    os.path.join(FETCHAI_PREF, "agents", "my_first_aea", AGENT_YAML),
    os.path.join(FETCHAI_PREF, "agents", "simple_service_registration", AGENT_YAML),
    os.path.join(FETCHAI_PREF, "agents", "tac_controller", AGENT_YAML),
    os.path.join(FETCHAI_PREF, "agents", "tac_controller_contract", AGENT_YAML),
    os.path.join(FETCHAI_PREF, "agents", "tac_participant", AGENT_YAML),
    os.path.join(FETCHAI_PREF, "agents", "thermometer_aea", AGENT_YAML),
    os.path.join(FETCHAI_PREF, "agents", "thermometer_client", AGENT_YAML),
    os.path.join(FETCHAI_PREF, "agents", "weather_client", AGENT_YAML),
    os.path.join(FETCHAI_PREF, "agents", "weather_station", AGENT_YAML),
]

protocol_specification_files = [
    os.path.join(PROTOCOL_SPECS_PREF_1, "sample.yaml",),
    os.path.join(PROTOCOL_SPECS_PREF_2, "sample_specification.yaml",),
    os.path.join(PROTOCOL_SPECS_PREF_2, "sample_specification_no_custom_types.yaml",),
]


def only_windows(fn: Callable) -> Callable:
    """
    Decorate a pytest method to run a test only in a case we are on Windows.

    :return: decorated method.
    """
    return action_for_platform("Windows", skip=False)(fn)


def skip_test_windows(fn: Callable) -> Callable:
    """
    Decorate a pytest method to skip a test in a case we are on Windows.

    :return: decorated method.
    """
    return action_for_platform("Windows", skip=True)(fn)


def skip_test_macos(fn: Callable) -> Callable:
    """
    Decorate a pytest method to skip a test in a case we are on MacOS.

    :return: decorated method.
    """
    return action_for_platform("Darwin", skip=True)(fn)


def action_for_platform(platform_name: str, skip: bool = True) -> Callable:
    """
    Decorate a pytest class or method to skip on certain platform.

    :param platform_name: check `platform.system()` for available platforms.
    :param skip: if True, the test will be skipped;
      if False, the test will be run ONLY on the chosen platform.

    :return: decorated object
    """
    # for docstyle.
    def decorator(pytest_func):
        """
        For the sake of clarity, assume the chosen platform for the action is "Windows".

        If the following condition is true:
          - the current system is not Windows (is_different) AND we want to skip it (skip)
         OR
          - the current system is Windows (not is_different) AND we want to run only on it (not skip)
        we run the test, else we skip the test.

        logically, the condition is a boolean equivalence
        between the variables "is_different" and "skip"
        Hence, the condition becomes:
        """
        is_different = platform.system() != platform_name
        if is_different is skip:
            return pytest_func

        def action(*args, **kwargs):
            if skip:
                pytest.skip(
                    f"Skipping the test since it doesn't work on {platform_name}."
                )
            else:
                pytest.skip(
                    f"Skipping the test since it works only on {platform_name}."
                )

        if isinstance(pytest_func, type):
            return type(
                pytest_func.__name__,
                (pytest_func,),
                {
                    "setup_class": action,
                    "setup": action,
                    "setUp": action,
                    "_skipped": True,
                },
            )

        @wraps(pytest_func)
        def wrapper(*args, **kwargs):  # type: ignore
            action(*args, **kwargs)

        return wrapper

    return decorator


@pytest.fixture(scope="session")
def oef_addr() -> str:
    """IP address pointing to the OEF Node to use during the tests."""
    return "127.0.0.1"


@pytest.fixture(scope="session")
def oef_port() -> int:
    """Port of the connection to the OEF Node to use during the tests."""
    return 10000


def tcpping(ip, port, log_exception: bool = True) -> bool:
    """Ping TCP port."""
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((ip, int(port)))
        s.shutdown(2)
        return True
    except Exception as e:
        if log_exception:
            logger.exception(e)
        return False


def wait_for_localhost_ports_to_close(
    ports: List[int], timeout: int = 120, sleep_time: int = 2
) -> None:
    """Wait for ports to close with timeout."""
    open_ports = ports
    elapsed = 0
    while len(open_ports) > 0 and elapsed < timeout:
        closed = []
        for port in open_ports:
            if not tcpping("127.0.0.1", port, log_exception=False):
                closed.append(port)
        open_ports = [port for port in open_ports if port not in closed]
        if len(open_ports) > 0:
            time.sleep(sleep_time)
            elapsed += sleep_time
    if open_ports != []:
        raise ValueError("Some ports are open: {}!".format(open_ports))


class OEFHealthCheck(object):
    """A health check class."""

    def __init__(
        self,
        oef_addr: str,
        oef_port: int,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        """
        Initialize.

        :param oef_addr: IP address of the OEF node.
        :param oef_port: Port of the OEF node.
        """
        self.oef_addr = oef_addr
        self.oef_port = oef_port

        self._result = False
        self._stop = False
        self._core = AsyncioCore()
        self.agent = OEFAgent(
            "check", core=self._core, oef_addr=self.oef_addr, oef_port=self.oef_port
        )
        self.agent.on_connect_success = self.on_connect_ok
        self.agent.on_connection_terminated = self.on_connect_terminated
        self.agent.on_connect_failed = self.exception_handler

    def exception_handler(self, url=None, ex=None):
        """Handle exception during a connection attempt."""
        print("An error occurred. Exception: {}".format(ex))
        self._stop = True

    def on_connect_ok(self, url=None):
        """Handle a successful connection."""
        print("Connection OK!")
        self._result = True
        self._stop = True

    def on_connect_terminated(self, url=None):
        """Handle a connection failure."""
        print("Connection terminated.")
        self._stop = True

    def run(self) -> bool:
        """
        Run the check, asynchronously.

        :return: True if the check is successful, False otherwise.
        """
        self._result = False
        self._stop = False

        def stop_connection_attempt(self):
            if self.agent.state == "connecting":
                self.agent.state = "failed"

        t = Timer(1.5, stop_connection_attempt, args=(self,))

        try:
            print("Connecting to {}:{}...".format(self.oef_addr, self.oef_port))
            self._core.run_threaded()

            t.start()
            self._result = self.agent.connect()
            self._stop = True

            if self._result:
                print("Connection established. Tearing down connection...")
                self.agent.disconnect()
                t.cancel()
            else:
                print("A problem occurred. Exiting...")
            return self._result

        except Exception as e:
            print(str(e))
            return self._result
        finally:
            t.join(1.0)
            self.agent.stop()
            self.agent.disconnect()
            self._core.stop()


def _stop_oef_search_images():
    """Stop the OEF search image."""
    client = docker.from_env()
    for container in client.containers.list():
        if "fetchai/oef-search:0.7" in container.image.tags:
            logger.info("Stopping existing Docker image...")
            container.stop()


def _wait_for_oef(max_attempts: int = 15, sleep_rate: float = 1.0):
    """Wait until the OEF is up."""
    success = False
    attempt = 0
    while not success and attempt < max_attempts:
        attempt += 1
        logger.info("Attempt {}...".format(attempt))
        oef_healthcheck = OEFHealthCheck("127.0.0.1", 10000)
        result = oef_healthcheck.run()
        if result:
            success = True
        else:
            logger.info(
                "OEF not available yet - sleeping for {} second...".format(sleep_rate)
            )
            time.sleep(sleep_rate)

    return success


def _create_oef_docker_image(oef_addr_, oef_port_) -> Container:
    client = docker.from_env()

    logger.info(ROOT_DIR + "/tests/common/oef_search_pluto_scripts")
    ports = {
        "20000/tcp": ("0.0.0.0", 20000),  # nosec
        "30000/tcp": ("0.0.0.0", 30000),  # nosec
        "{}/tcp".format(oef_port_): ("0.0.0.0", oef_port_),  # nosec
    }
    volumes = {
        ROOT_DIR
        + "/tests/common/oef_search_pluto_scripts": {"bind": "/config", "mode": "rw"},
        ROOT_DIR + "/data/oef-logs": {"bind": "/logs", "mode": "rw"},
    }
    c = client.containers.run(
        "fetchai/oef-search:0.7",
        "/config/node_config.json",
        detach=True,
        ports=ports,
        volumes=volumes,
    )
    return c


def pytest_addoption(parser) -> None:
    """Add --aea-loop option."""
    parser.addoption(
        "--aea-loop",
        action="store",
        default="async",
        help="aea loop to use: async[default] or sync",
    )
    # disable inernet connection
    parser.addoption(
        "--no-inet",
        action="store_true",
        default=False,
        help="block socket connect outside of 127.x.x.x",
    )

    parser.addoption(
        "--check-threads",
        action="store_true",
        default=False,
        help="check non closed threads i started during test",
    )


@pytest.fixture(scope="session", autouse=True)
def inet_disable(request) -> None:
    """Disable internet access via socket."""
    if not request.config.getoption("--no-inet"):
        return

    orig_connect = socket.socket.connect

    def socket_connect(*args):
        host = args[1][0]
        if host == "localhost" or host.startswith("127."):
            return orig_connect(*args)
        raise socket.error("Internet disabled by pytest option --no-inet")

    p = patch.object(socket.socket, "connect", new=socket_connect)
    p.start()


@pytest.fixture(scope="session", autouse=True)
def apply_aea_loop(request) -> None:
    """Patch AEA.DEFAULT_RUN_LOOP using pytest option `--aea-loop`."""
    loop = request.config.getoption("--aea-loop")
    assert loop in AEA.RUN_LOOPS
    AEA.DEFAULT_RUN_LOOP = loop


@pytest.fixture(scope="session")
def network_node(
    oef_addr, oef_port, pytestconfig, timeout: float = 2.0, max_attempts: int = 10
):
    """Network node initialization."""
    if sys.version_info < (3, 7):
        pytest.skip("Python version < 3.7 not supported by the OEF.")
        return

    if os.name == "nt":
        pytest.skip("Skip test as it doesn't work on Windows.")

    _stop_oef_search_images()
    c = _create_oef_docker_image(oef_addr, oef_port)
    c.start()

    # wait for the setup...
    logger.info("Setting up the OEF node...")
    success = _wait_for_oef(max_attempts=max_attempts, sleep_rate=timeout)

    if not success:
        c.stop()
        c.remove()
        pytest.fail("OEF doesn't work. Exiting...")
    else:
        logger.info("Done!")
        time.sleep(timeout)
        yield
        logger.info("Stopping the OEF node...")
        c.stop()
        c.remove()


@pytest.fixture(scope="session", autouse=True)
def reset_aea_cli_config() -> None:
    """Reset the cli config for each test."""
    _init_cli_config()


def get_unused_tcp_port():
    """Get an unused TCP port."""
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()
    return port


def get_host():
    """Get the host."""
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(("10.255.255.255", 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = "127.0.0.1"
    finally:
        s.close()
    return IP


def double_escape_windows_path_separator(path):
    r"""Doubleescape Windows path separator '\'."""
    return path.replace("\\", "\\\\")


def _make_dummy_connection() -> Connection:
    configuration = ConnectionConfig(connection_id=DummyConnection.connection_id,)
    dummy_connection = DummyConnection(
        configuration=configuration, identity=Identity("name", "address")
    )
    return dummy_connection


def _make_local_connection(
    address: Address,
    node: LocalNode,
    restricted_to_protocols=None,
    excluded_protocols=None,
) -> Connection:
    configuration = ConnectionConfig(
        restricted_to_protocols=restricted_to_protocols,
        excluded_protocols=excluded_protocols,
        connection_id=OEFLocalConnection.connection_id,
    )
    oef_local_connection = OEFLocalConnection(
        configuration=configuration, identity=Identity("name", address), local_node=node
    )
    return oef_local_connection


def _make_oef_connection(address: Address, oef_addr: str, oef_port: int):
    configuration = ConnectionConfig(
        addr=oef_addr, port=oef_port, connection_id=OEFConnection.connection_id
    )
    oef_connection = OEFConnection(
        configuration=configuration, identity=Identity("name", address),
    )
    oef_connection._default_logger_name = "aea.packages.fetchai.connections.oef"
    return oef_connection


def _make_tcp_server_connection(address: str, host: str, port: int):
    configuration = ConnectionConfig(
        address=host, port=port, connection_id=TCPServerConnection.connection_id
    )
    tcp_connection = TCPServerConnection(
        configuration=configuration, identity=Identity("name", address),
    )
    tcp_connection._default_logger_name = (
        "aea.packages.fetchai.connections.tcp.tcp_server"
    )
    return tcp_connection


def _make_tcp_client_connection(address: str, host: str, port: int):
    configuration = ConnectionConfig(
        address=host, port=port, connection_id=TCPClientConnection.connection_id
    )
    tcp_connection = TCPClientConnection(
        configuration=configuration, identity=Identity("name", address),
    )
    tcp_connection._default_logger_name = (
        "aea.packages.fetchai.connections.tcp.tcp_client"
    )
    return tcp_connection


def _make_stub_connection(input_file_path: str, output_file_path: str):
    configuration = ConnectionConfig(
        input_file=input_file_path,
        output_file=output_file_path,
        connection_id=StubConnection.connection_id,
    )
    connection = StubConnection(configuration=configuration)
    return connection


def _make_libp2p_connection(
    port: int = 10234,
    host: str = "127.0.0.1",
    relay: bool = True,
    delegate: bool = False,
    entry_peers: Optional[Sequence[MultiAddr]] = None,
    delegate_port: int = 11234,
    delegate_host: str = "127.0.0.1",
    node_key_file: Optional[str] = None,
    agent_address: Optional[Address] = None,
) -> P2PLibp2pConnection:
    log_file = "libp2p_node_{}.log".format(port)
    if os.path.exists(log_file):
        os.remove(log_file)
    address = agent_address
    if address is None:
        address = make_crypto(COSMOS).address
    identity = Identity("", address=address)
    if relay and delegate:
        configuration = ConnectionConfig(
            node_key_file=node_key_file,
            local_uri="{}:{}".format(host, port),
            public_uri="{}:{}".format(host, port),
            entry_peers=entry_peers,
            log_file=log_file,
            delegate_uri="{}:{}".format(delegate_host, delegate_port),
            connection_id=P2PLibp2pConnection.connection_id,
        )
    elif relay and not delegate:
        configuration = ConnectionConfig(
            node_key_file=node_key_file,
            local_uri="{}:{}".format(host, port),
            public_uri="{}:{}".format(host, port),
            entry_peers=entry_peers,
            log_file=log_file,
            connection_id=P2PLibp2pConnection.connection_id,
        )
    else:
        configuration = ConnectionConfig(
            node_key_file=node_key_file,
            local_uri="{}:{}".format(host, port),
            entry_peers=entry_peers,
            log_file=log_file,
            connection_id=P2PLibp2pConnection.connection_id,
        )
    return P2PLibp2pConnection(configuration=configuration, identity=identity)


def _make_libp2p_client_connection(
    node_port: int = 11234, node_host: str = "127.0.0.1", uri: Optional[str] = None,
) -> P2PLibp2pClientConnection:
    crypto = make_crypto(COSMOS)
    identity = Identity("", address=crypto.address)
    configuration = ConnectionConfig(
        client_key_file=None,
        nodes=[
            {
                "uri": str(uri)
                if uri is not None
                else "{}:{}".format(node_host, node_port)
            }
        ],
        connection_id=P2PLibp2pClientConnection.connection_id,
    )
    return P2PLibp2pClientConnection(configuration=configuration, identity=identity)


def libp2p_log_on_failure(fn: Callable) -> Callable:
    """
    Decorate a pytest method running a libp2p node to print its logs in case test fails.

    :return: decorated method.
    """
    # for pydcostyle
    @wraps(fn)
    def wrapper(self, *args, **kwargs):
        try:
            fn(self, *args, **kwargs)
        except Exception as e:
            for log_file in self.log_files:
                print("libp2p log file ======================= {}".format(log_file))
                with open(log_file, "r") as f:
                    print(f.read())
                print("=======================================")
            raise e

    return wrapper


def libp2p_log_on_failure_all(cls):
    """
    Decorate every method of a class with `libp2p_log_on_failure`.

    :return: class with decorated methods.
    """
    # TODO(LR) test it is a type
    for name, fn in inspect.getmembers(cls):
        if isinstance(fn, FunctionType):
            setattr(cls, name, libp2p_log_on_failure(fn))
        # TOFIX(LR) decorate already @classmethod decorated methods
        continue
        if isinstance(fn, MethodType):
            if fn.im_self is None:
                wrapped_fn = libp2p_log_on_failure(fn.im_func)
                method = MethodType(wrapped_fn, None, cls)
                setattr(cls, name, method)
            else:
                wrapped_fn = libp2p_log_on_failure(fn.im_func)
                clsmethod = MethodType(wrapped_fn, cls, type)
                setattr(cls, name, clsmethod)
    return cls


def _do_for_all(method_decorator):
    def class_decorator(cls):
        class GetAttributeMetaClass(type):
            def __getattribute__(cls, name):
                attr = super().__getattribute__(name)
                return method_decorator(attr)

        class DecoratedClass(cls, metaclass=GetAttributeMetaClass):
            def __getattribute__(self, name):
                attr = super().__getattribute__(name)
                return method_decorator(attr)

        for attr in WRAPPER_ASSIGNMENTS:
            if not hasattr(cls, attr):
                continue
            setattr(DecoratedClass, attr, getattr(cls, attr))
        DecoratedClass.__wrapped__ = cls
        return DecoratedClass

    return class_decorator


class CwdException(Exception):
    """Exception to raise if cwd was not restored by test."""

    def __init__(self):
        """Init expcetion with default message."""
        super().__init__("CWD was not restored")


@pytest.fixture(scope="class", autouse=True)
def aea_testcase_teardown_check(request):
    """Check BaseAEATestCase.teardown_class for BaseAEATestCase based test cases."""
    from aea.test_tools.test_cases import BaseAEATestCase  # cause circular import

    yield
    if (
        request.cls
        and issubclass(request.cls, BaseAEATestCase)
        and getattr(request.cls, "_skipped", False) is False
    ):
        assert getattr(
            request.cls, "_is_teardown_class_called", None
        ), "No BaseAEATestCase.teardown_class was called!"


@pytest.fixture(scope="class", autouse=True)
def check_test_class_cwd():
    """Check test case class restore CWD."""
    os.chdir(ROOT_DIR)
    old_cwd = os.getcwd()
    yield
    if old_cwd != os.getcwd():
        raise CwdException()


@pytest.fixture(autouse=True)
def check_test_cwd(request):
    """Check particular test restore CWD."""
    if request.cls:
        yield
        return
    os.chdir(ROOT_DIR)
    old_cwd = os.getcwd()
    yield
    if old_cwd != os.getcwd():
        os.chdir(ROOT_DIR)
        raise CwdException()


@pytest.fixture(autouse=True)
def set_logging_to_debug(request):
    """Set aea logger to debug."""
    aea_logger = logging.getLogger("aea")
    aea_logger.setLevel(logging.DEBUG)


@pytest.fixture(autouse=True)
def check_test_threads(request):
    """Check particular test close all spawned threads."""
    if not request.config.getoption("--check-threads"):
        yield
        return
    if request.cls:
        yield
        return
    num_threads = threading.activeCount()
    yield
    new_num_threads = threading.activeCount()
    assert num_threads >= new_num_threads, "Non closed threads!"


@pytest.fixture()
async def ledger_apis_connection(request):
    """Make a connection."""
    crypto = make_crypto(DEFAULT_LEDGER)
    identity = Identity("name", crypto.address)
    crypto_store = CryptoStore()
    directory = Path(ROOT_DIR, "packages", "fetchai", "connections", "ledger")
    connection = Connection.from_dir(
        directory, identity=identity, crypto_store=crypto_store
    )
    connection = cast(Connection, connection)
    connection._logger = logging.getLogger("aea.packages.fetchai.connections.ledger")
    await connection.connect()
    yield connection
    await connection.disconnect()


@pytest.fixture()
def erc1155_contract():
    """
    Instantiate an ERC1155 contract instance.

    As a side effect, register it to the registry, if not already registered.
    """
    directory = Path(ROOT_DIR, "packages", "fetchai", "contracts", "erc1155")
    configuration = load_component_configuration(ComponentType.CONTRACT, directory)
    configuration._directory = directory
    configuration = cast(ContractConfig, configuration)

    if str(configuration.public_id) not in contract_registry.specs:
        # load contract into sys modules
        Contract.from_config(configuration)

    contract = contract_registry.make(str(configuration.public_id))
    yield contract


def env_path_separator() -> str:
    """
    Get the separator between path items in PATH variables, cross platform.

    E.g. on Linux and MacOS, it returns ':', whereas on Windows ';'.
    """
    if sys.platform == "win32":
        return ";"
    else:
        return ":"



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains tests for aea runtime."""
import asyncio
import os
import time
from pathlib import Path
from typing import Type
from unittest.mock import patch

import pytest

from aea.aea_builder import AEABuilder
from aea.configurations.constants import DEFAULT_LEDGER, DEFAULT_PRIVATE_KEY_FILE
from aea.runtime import (
    AsyncRuntime,
    BaseRuntime,
    RuntimeStates,
    ThreadedRuntime,
    _StopRuntime,
)

from tests.common.utils import wait_for_condition
from tests.conftest import CUR_PATH, DUMMY_SKILL_PUBLIC_ID


class TestAsyncRuntime:
    """Test async runtime."""

    RUNTIME: Type[BaseRuntime] = AsyncRuntime

    def setup(self):
        """Set up case."""
        agent_name = "MyAgent"
        private_key_path = os.path.join(CUR_PATH, "data", DEFAULT_PRIVATE_KEY_FILE)
        builder = AEABuilder()
        builder.set_name(agent_name)
        builder.add_private_key(DEFAULT_LEDGER, private_key_path)
        builder.add_skill(Path(CUR_PATH, "data", "dummy_skill"))
        self.agent = builder.build()
        self.runtime = self.RUNTIME(self.agent, threaded=True)

    def teardown(self):
        """Tear down."""
        self.runtime.stop()
        self.runtime.wait_completed(sync=True)

    def test_start_stop(self):
        """Test runtime tart/stop."""
        self.runtime.start()
        wait_for_condition(lambda: self.runtime.is_running, timeout=20)
        self.runtime.stop()
        self.runtime.wait_completed(sync=True)

    def test_stop_with_stopped_exception(self):
        """Test runtime stopped by stopruntime exception."""
        behaviour = self.agent.resources.get_behaviour(DUMMY_SKILL_PUBLIC_ID, "dummy")
        with patch.object(
            behaviour, "act", side_effect=_StopRuntime(reraise=ValueError("expected"))
        ):
            self.runtime.start()
            wait_for_condition(lambda: self.runtime.is_running, timeout=20)
            time.sleep(1)
            assert self.runtime.is_stopped

        with pytest.raises(ValueError, match="expected"):
            self.runtime.wait_completed(timeout=10, sync=True)

    def test_double_start(self):
        """Test runtime double start do nothing."""
        assert self.runtime.start()
        assert not self.runtime.start()
        wait_for_condition(lambda: self.runtime.is_running, timeout=20)

    def test_set_loop(self):
        """Test set loop method."""
        loop = asyncio.new_event_loop()
        self.runtime.set_loop(loop)
        assert self.runtime.loop is loop

    def test_double_stop(self):
        """Test runtime double stop do nothing."""
        self.runtime.start()
        wait_for_condition(lambda: self.runtime.is_running, timeout=20)
        self.runtime.stop()
        self.runtime.wait_completed(sync=True)
        assert self.runtime.is_stopped

        self.runtime.stop()
        self.runtime.wait_completed(sync=True)
        assert self.runtime.is_stopped

    def test_error_state(self):
        """Test runtime fails on start."""
        with patch.object(
            self.runtime, "_start_agent_loop", side_effect=ValueError("oops")
        ):
            with pytest.raises(ValueError, match="oops"):
                self.runtime.start_and_wait_completed(sync=True)

        assert self.runtime.state == RuntimeStates.error, self.runtime.state


class TestThreadedRuntime(TestAsyncRuntime):
    """Test threaded runtime."""

    RUNTIME = ThreadedRuntime

    def test_error_state(self):
        """Test runtime fails on start."""
        with patch.object(
            self.runtime.main_loop, "start", side_effect=ValueError("oops")
        ):
            with pytest.raises(ValueError, match="oops"):
                self.runtime.start_and_wait_completed(sync=True)

        assert self.runtime.state == RuntimeStates.error, self.runtime.state



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This tool generates the API docs."""

import shutil
import subprocess  # nosec
import sys
from pathlib import Path


DOCS_DIR = "docs/"
MODULES_TO_PATH = {
    "aea.abstract_agent": "api/abstract_agent.md",
    "aea.aea": "api/aea.md",
    "aea.aea_builder": "api/aea_builder.md",
    "aea.agent": "api/agent.md",
    "aea.agent_loop": "api/agent_loop.md",
    "aea.common": "api/common.md",
    "aea.exceptions": "api/exceptions.md",
    "aea.launcher": "api/launcher.md",
    "aea.manager": "api/manager.md",
    "aea.multiplexer": "api/multiplexer.md",
    "aea.runner": "api/runner.md",
    "aea.runtime": "api/runtime.md",
    "aea.components.base": "api/components/base.md",
    "aea.components.loader": "api/components/loader.md",
    "aea.configurations.base": "api/configurations/base.md",
    "aea.configurations.constants": "api/configurations/constants.md",
    "aea.configurations.loader": "api/configurations/loader.md",
    "aea.configurations.project": "api/configurations/project.md",
    "aea.configurations.pypi": "api/configurations/pypi.md",
    "aea.connections.base": "api/connections/base.md",
    "aea.connections.stub.connection": "api/connections/stub/connection.md",
    "aea.context.base": "api/context/base.md",
    "aea.contracts.base": "api/contracts/base.md",
    "aea.crypto.base": "api/crypto/base.md",
    "aea.crypto.cosmos": "api/crypto/cosmos.md",
    "aea.crypto.ethereum": "api/crypto/ethereum.md",
    "aea.crypto.fetchai": "api/crypto/fetchai.md",
    "aea.crypto.helpers": "api/crypto/helpers.md",
    "aea.crypto.ledger_apis": "api/crypto/ledger_apis.md",
    "aea.crypto.wallet": "api/crypto/wallet.md",
    "aea.crypto.registries.base": "api/crypto/registries/base.md",
    "aea.decision_maker.base": "api/decision_maker/base.md",
    "aea.decision_maker.default": "api/decision_maker/default.md",
    "aea.helpers.ipfs.base": "api/helpers/ipfs/base.md",
    "aea.helpers.multiaddr.base": "api/helpers/multiaddr/base.md",
    "aea.helpers.preference_representations.base": "api/helpers/preference_representations/base.md",
    "aea.helpers.search.generic": "api/helpers/search/generic.md",
    "aea.helpers.search.models": "api/helpers/search/models.md",
    "aea.helpers.transaction.base": "api/helpers/transaction/base.md",
    "aea.helpers.async_friendly_queue": "api/helpers/async_friendly_queue.md",
    "aea.helpers.async_utils": "api/helpers/async_utils.md",
    "aea.helpers.base": "api/helpers/base.md",
    "aea.helpers.exception_policy": "api/helpers/exception_policy.md",
    "aea.helpers.exec_timeout": "api/helpers/exec_timeout.md",
    "aea.helpers.file_lock": "api/helpers/file_lock.md",
    "aea.helpers.logging": "api/helpers/logging.md",
    "aea.helpers.multiple_executor": "api/helpers/multiple_executor.md",
    "aea.helpers.pipe": "api/helpers/pipe.md",
    "aea.helpers.win32": "api/helpers/win32.md",
    "aea.helpers.yaml_utils": "api/helpers/yaml_utils.md",
    "aea.identity.base": "api/identity/base.md",
    "aea.mail.base": "api/mail/base.md",
    "aea.protocols.base": "api/protocols/base.md",
    "aea.protocols.dialogue.base": "api/helpers/dialogue/base.md",
    "aea.protocols.generator.base": "api/protocols/generator/base.md",
    "aea.protocols.generator.common": "api/protocols/generator/common.md",
    "aea.protocols.generator.extract_specification": "api/protocols/generator/extract_specification.md",
    "aea.protocols.generator.validate": "api/protocols/generator/validate.md",
    "aea.protocols.default.custom_types": "api/protocols/default/custom_types.md",
    "aea.protocols.default.dialogues": "api/protocols/default/dialogues.md",
    "aea.protocols.default.message": "api/protocols/default/message.md",
    "aea.protocols.default.serialization": "api/protocols/default/serialization.md",
    "aea.protocols.signing.custom_types": "api/protocols/signing/custom_types.md",
    "aea.protocols.signing.dialogues": "api/protocols/signing/dialogues.md",
    "aea.protocols.signing.message": "api/protocols/signing/message.md",
    "aea.protocols.signing.serialization": "api/protocols/signing/serialization.md",
    "aea.protocols.state_update.dialogues": "api/protocols/state_update/dialogues.md",
    "aea.protocols.state_update.message": "api/protocols/state_update/message.md",
    "aea.protocols.state_update.serialization": "api/protocols/state_update/serialization.md",
    "aea.registries.base": "api/registries/base.md",
    "aea.registries.filter": "api/registries/filter.md",
    "aea.registries.resources": "api/registries/resources.md",
    "aea.skills.base": "api/skills/base.md",
    "aea.skills.behaviours": "api/skills/behaviours.md",
    "aea.skills.tasks": "api/skills/tasks.md",
    "aea.skills.error.handlers": "api/skills/error/handlers.md",
    "aea.test_tools.generic": "api/test_tools/generic.md",
    "aea.test_tools.test_cases": "api/test_tools/test_cases.md",
    "aea.test_tools.test_skill": "api/test_tools/test_skill.md",
}


def create_subdir(path) -> None:
    """
    Create a subdirectory.

    :param path: the directory path
    """
    directory = "/".join(path.split("/")[:-1])
    Path(directory).mkdir(parents=True, exist_ok=True)


def replace_underscores(text: str) -> str:
    """
    Replace escaped underscores in a text.

    :return: the processed text
    """
    text_a = text.replace("\\_\\_", "`__`")
    text_b = text_a.replace("\\_", "`_`")
    return text_b


def save_to_file(path: str, text: str) -> None:
    """
    Save to a file path.

    :param path: the path
    :param text: the text
    :return: None
    """
    with open(path, "w") as f:
        f.write(text)


def generate_api_docs():
    """Generate the api docs."""
    for module, rel_path in MODULES_TO_PATH.items():
        path = DOCS_DIR + rel_path
        create_subdir(path)
        pydoc = subprocess.Popen(  # nosec
            ["pydoc-markdown", "-m", module, "-I", "."], stdout=subprocess.PIPE
        )
        stdout, _ = pydoc.communicate()
        pydoc.wait()
        stdout_text = stdout.decode("utf-8")
        text = replace_underscores(stdout_text)
        save_to_file(path, text)


def install(package: str) -> int:
    """
    Install a PyPI package by calling pip.

    :param package: the package name and version specifier.
    :return: the return code.
    """
    return subprocess.check_call(  # nosec
        [sys.executable, "-m", "pip", "install", package]
    )


if __name__ == "__main__":
    res = shutil.which("pydoc-markdown")
    if res is None:
        install("pydoc-markdown==3.3.0")
        sys.exit(1)
    generate_api_docs()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for AsyncFriendlyQueue."""

import asyncio
import time
from queue import Empty
from threading import Thread

import pytest

from aea.helpers.async_friendly_queue import AsyncFriendlyQueue


def test_same_thread() -> None:
    """Test AsyncFriendlyQueue in one thread environment."""
    sq = AsyncFriendlyQueue()

    with pytest.raises(Empty):
        sq.get_nowait()

    item = "item"
    sq.put_nowait(item)
    assert sq.get_nowait() == item


@pytest.mark.asyncio
async def test_asyncio_loop() -> None:
    """Test AsyncFriendlyQueue inside one event loop."""
    sq = AsyncFriendlyQueue()
    item = "item"
    sq.put(item)
    assert await sq.async_get() == item


def test_many_threads_with_asyncio() -> None:
    """Test AsyncFriendlyQueue wuth multiple asyncio event loop consumers in different threads."""
    sq = AsyncFriendlyQueue()
    num_threads = 10
    threads = []
    results = []  # type: ignore

    def test(sq, timeout, results):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        async def wait_msg(sq, timeout, results):
            await asyncio.wait_for(sq.async_get(), timeout)
            results.append("done")

        loop.run_until_complete(wait_msg(sq, timeout, results))
        loop.close()

    for _ in range(num_threads):
        t = Thread(target=test, args=(sq, 5, results))
        t.daemon = True
        t.start()
        threads.append(t)

    time.sleep(0.03)
    for _ in range(num_threads):
        sq.put("item")

    for t in threads:
        t.join()

    assert len(results) == num_threads



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Implementation of the 'aea delete' subcommand."""

import os
import shutil
from typing import cast

import click

from aea.cli.utils.click_utils import AgentDirectory
from aea.cli.utils.context import Context


@click.command()
@click.argument(
    "agent_name", type=AgentDirectory(), required=True,
)
@click.pass_context
def delete(click_context, agent_name):
    """Delete an agent."""
    click.echo("Deleting AEA project directory './{}'...".format(agent_name))
    ctx = cast(Context, click_context.obj)
    delete_aea(ctx, agent_name)


def delete_aea(ctx: Context, agent_name: str) -> None:
    """
    Delete agent's directory.

    :param agent_name: name of the agent (equal to folder name).

    :return: None
    :raises: ClickException if OSError occurred.
    """
    agent_path = os.path.join(ctx.cwd, agent_name)
    try:
        shutil.rmtree(agent_path, ignore_errors=False)
    except OSError:
        raise click.ClickException(
            "An error occurred while deleting the agent directory. Aborting..."
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains the tests for commands in aea.cli.generate_wealth module."""

from unittest import TestCase, mock

import pytest

from aea.cli import cli
from aea.cli.generate_wealth import _try_generate_wealth
from aea.test_tools.exceptions import AEATestingException
from aea.test_tools.test_cases import AEATestCaseMany

from tests.conftest import (
    CLI_LOG_OPTION,
    CliRunner,
    FETCHAI,
    MAX_FLAKY_RERUNS_INTEGRATION,
)
from tests.test_cli.tools_for_testing import ContextMock


class GenerateWealthTestCase(TestCase):
    """Test case for _generate_wealth method."""

    @mock.patch("aea.cli.utils.package_utils.Wallet")
    @mock.patch("aea.cli.generate_wealth.click.echo")
    @mock.patch("aea.cli.generate_wealth.try_generate_testnet_wealth")
    @mock.patch("aea.cli.utils.package_utils.verify_or_create_private_keys_ctx")
    def test__generate_wealth_positive(self, *mocks):
        """Test for _generate_wealth method positive result."""
        ctx = ContextMock()
        _try_generate_wealth(ctx, "cosmos", True)


@mock.patch("aea.cli.utils.decorators.try_to_load_agent_config")
@mock.patch("aea.cli.utils.package_utils.verify_or_create_private_keys_ctx")
@mock.patch("aea.cli.generate_wealth._try_generate_wealth")
class GenerateWealthCommandTestCase(TestCase):
    """Test case for CLI generate_wealth command."""

    def setUp(self):
        """Set it up."""
        self.runner = CliRunner()

    def test_run_positive(self, *mocks):
        """Test for CLI generate_wealth positive result."""
        result = self.runner.invoke(
            cli,
            [
                *CLI_LOG_OPTION,
                "--skip-consistency-check",
                "generate-wealth",
                "--sync",
                FETCHAI,
            ],
            standalone_mode=False,
        )
        self.assertEqual(result.exit_code, 0)


class TestWealthCommandsPositive(AEATestCaseMany):
    """Test case for CLI wealth commands."""

    @pytest.mark.integration
    @pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS_INTEGRATION)
    def test_wealth_commands(self):
        """Test wealth commands."""
        agent_name = "test_aea"
        self.create_agents(agent_name)

        self.set_agent_context(agent_name)

        self.generate_private_key()
        self.add_private_key()

        self.generate_wealth()


class TestWealthCommandsNegative(AEATestCaseMany):
    """Test case for CLI wealth commands, negative case."""

    def test_wealth_commands_negative(self):
        """Test wealth commands."""
        agent_name = "test_aea"
        self.create_agents(agent_name)

        self.set_agent_context(agent_name)

        self.generate_private_key()
        self.add_private_key()

        settings = {"unsupported_crypto": "path"}
        self.nested_set_config("agent.private_key_paths", settings)
        with pytest.raises(AEATestingException) as excinfo:
            self.generate_wealth()

        assert "Item not registered with id 'unsupported_crypto'." in str(excinfo.value)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: fipa.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="fipa.proto",
    package="fetch.aea.Fipa",
    syntax="proto3",
    serialized_options=None,
    serialized_pb=b'\n\nfipa.proto\x12\x0e\x66\x65tch.aea.Fipa"\xc4\x0c\n\x0b\x46ipaMessage\x12\x12\n\nmessage_id\x18\x01 \x01(\x05\x12"\n\x1a\x64ialogue_starter_reference\x18\x02 \x01(\t\x12$\n\x1c\x64ialogue_responder_reference\x18\x03 \x01(\t\x12\x0e\n\x06target\x18\x04 \x01(\x05\x12\x41\n\x06\x61\x63\x63\x65pt\x18\x05 \x01(\x0b\x32/.fetch.aea.Fipa.FipaMessage.Accept_PerformativeH\x00\x12S\n\x0f\x61\x63\x63\x65pt_w_inform\x18\x06 \x01(\x0b\x32\x38.fetch.aea.Fipa.FipaMessage.Accept_W_Inform_PerformativeH\x00\x12;\n\x03\x63\x66p\x18\x07 \x01(\x0b\x32,.fetch.aea.Fipa.FipaMessage.Cfp_PerformativeH\x00\x12\x43\n\x07\x64\x65\x63line\x18\x08 \x01(\x0b\x32\x30.fetch.aea.Fipa.FipaMessage.Decline_PerformativeH\x00\x12\x41\n\x06inform\x18\t \x01(\x0b\x32/.fetch.aea.Fipa.FipaMessage.Inform_PerformativeH\x00\x12M\n\x0cmatch_accept\x18\n \x01(\x0b\x32\x35.fetch.aea.Fipa.FipaMessage.Match_Accept_PerformativeH\x00\x12_\n\x15match_accept_w_inform\x18\x0b \x01(\x0b\x32>.fetch.aea.Fipa.FipaMessage.Match_Accept_W_Inform_PerformativeH\x00\x12\x43\n\x07propose\x18\x0c \x01(\x0b\x32\x30.fetch.aea.Fipa.FipaMessage.Propose_PerformativeH\x00\x1a"\n\x0b\x44\x65scription\x12\x13\n\x0b\x64\x65scription\x18\x01 \x01(\x0c\x1a\x81\x01\n\x05Query\x12\x0f\n\x05\x62ytes\x18\x01 \x01(\x0cH\x00\x12<\n\x07nothing\x18\x02 \x01(\x0b\x32).fetch.aea.Fipa.FipaMessage.Query.NothingH\x00\x12\x15\n\x0bquery_bytes\x18\x03 \x01(\x0cH\x00\x1a\t\n\x07NothingB\x07\n\x05query\x1a\x44\n\x10\x43\x66p_Performative\x12\x30\n\x05query\x18\x01 \x01(\x0b\x32!.fetch.aea.Fipa.FipaMessage.Query\x1aQ\n\x14Propose_Performative\x12\x39\n\x08proposal\x18\x01 \x01(\x0b\x32\'.fetch.aea.Fipa.FipaMessage.Description\x1a\x9d\x01\n\x1c\x41\x63\x63\x65pt_W_Inform_Performative\x12P\n\x04info\x18\x01 \x03(\x0b\x32\x42.fetch.aea.Fipa.FipaMessage.Accept_W_Inform_Performative.InfoEntry\x1a+\n\tInfoEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xa9\x01\n"Match_Accept_W_Inform_Performative\x12V\n\x04info\x18\x01 \x03(\x0b\x32H.fetch.aea.Fipa.FipaMessage.Match_Accept_W_Inform_Performative.InfoEntry\x1a+\n\tInfoEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x8b\x01\n\x13Inform_Performative\x12G\n\x04info\x18\x01 \x03(\x0b\x32\x39.fetch.aea.Fipa.FipaMessage.Inform_Performative.InfoEntry\x1a+\n\tInfoEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x15\n\x13\x41\x63\x63\x65pt_Performative\x1a\x16\n\x14\x44\x65\x63line_Performative\x1a\x1b\n\x19Match_Accept_PerformativeB\x0e\n\x0cperformativeb\x06proto3',
)


_FIPAMESSAGE_DESCRIPTION = _descriptor.Descriptor(
    name="Description",
    full_name="fetch.aea.Fipa.FipaMessage.Description",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="description",
            full_name="fetch.aea.Fipa.FipaMessage.Description.description",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=750,
    serialized_end=784,
)

_FIPAMESSAGE_QUERY_NOTHING = _descriptor.Descriptor(
    name="Nothing",
    full_name="fetch.aea.Fipa.FipaMessage.Query.Nothing",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=898,
    serialized_end=907,
)

_FIPAMESSAGE_QUERY = _descriptor.Descriptor(
    name="Query",
    full_name="fetch.aea.Fipa.FipaMessage.Query",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="bytes",
            full_name="fetch.aea.Fipa.FipaMessage.Query.bytes",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="nothing",
            full_name="fetch.aea.Fipa.FipaMessage.Query.nothing",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="query_bytes",
            full_name="fetch.aea.Fipa.FipaMessage.Query.query_bytes",
            index=2,
            number=3,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[_FIPAMESSAGE_QUERY_NOTHING,],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="query",
            full_name="fetch.aea.Fipa.FipaMessage.Query.query",
            index=0,
            containing_type=None,
            fields=[],
        ),
    ],
    serialized_start=787,
    serialized_end=916,
)

_FIPAMESSAGE_CFP_PERFORMATIVE = _descriptor.Descriptor(
    name="Cfp_Performative",
    full_name="fetch.aea.Fipa.FipaMessage.Cfp_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="query",
            full_name="fetch.aea.Fipa.FipaMessage.Cfp_Performative.query",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=918,
    serialized_end=986,
)

_FIPAMESSAGE_PROPOSE_PERFORMATIVE = _descriptor.Descriptor(
    name="Propose_Performative",
    full_name="fetch.aea.Fipa.FipaMessage.Propose_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="proposal",
            full_name="fetch.aea.Fipa.FipaMessage.Propose_Performative.proposal",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=988,
    serialized_end=1069,
)

_FIPAMESSAGE_ACCEPT_W_INFORM_PERFORMATIVE_INFOENTRY = _descriptor.Descriptor(
    name="InfoEntry",
    full_name="fetch.aea.Fipa.FipaMessage.Accept_W_Inform_Performative.InfoEntry",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="fetch.aea.Fipa.FipaMessage.Accept_W_Inform_Performative.InfoEntry.key",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="fetch.aea.Fipa.FipaMessage.Accept_W_Inform_Performative.InfoEntry.value",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b"8\001",
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1186,
    serialized_end=1229,
)

_FIPAMESSAGE_ACCEPT_W_INFORM_PERFORMATIVE = _descriptor.Descriptor(
    name="Accept_W_Inform_Performative",
    full_name="fetch.aea.Fipa.FipaMessage.Accept_W_Inform_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="info",
            full_name="fetch.aea.Fipa.FipaMessage.Accept_W_Inform_Performative.info",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[_FIPAMESSAGE_ACCEPT_W_INFORM_PERFORMATIVE_INFOENTRY,],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1072,
    serialized_end=1229,
)

_FIPAMESSAGE_MATCH_ACCEPT_W_INFORM_PERFORMATIVE_INFOENTRY = _descriptor.Descriptor(
    name="InfoEntry",
    full_name="fetch.aea.Fipa.FipaMessage.Match_Accept_W_Inform_Performative.InfoEntry",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="fetch.aea.Fipa.FipaMessage.Match_Accept_W_Inform_Performative.InfoEntry.key",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="fetch.aea.Fipa.FipaMessage.Match_Accept_W_Inform_Performative.InfoEntry.value",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b"8\001",
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1186,
    serialized_end=1229,
)

_FIPAMESSAGE_MATCH_ACCEPT_W_INFORM_PERFORMATIVE = _descriptor.Descriptor(
    name="Match_Accept_W_Inform_Performative",
    full_name="fetch.aea.Fipa.FipaMessage.Match_Accept_W_Inform_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="info",
            full_name="fetch.aea.Fipa.FipaMessage.Match_Accept_W_Inform_Performative.info",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[_FIPAMESSAGE_MATCH_ACCEPT_W_INFORM_PERFORMATIVE_INFOENTRY,],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1232,
    serialized_end=1401,
)

_FIPAMESSAGE_INFORM_PERFORMATIVE_INFOENTRY = _descriptor.Descriptor(
    name="InfoEntry",
    full_name="fetch.aea.Fipa.FipaMessage.Inform_Performative.InfoEntry",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="fetch.aea.Fipa.FipaMessage.Inform_Performative.InfoEntry.key",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="fetch.aea.Fipa.FipaMessage.Inform_Performative.InfoEntry.value",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b"8\001",
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1186,
    serialized_end=1229,
)

_FIPAMESSAGE_INFORM_PERFORMATIVE = _descriptor.Descriptor(
    name="Inform_Performative",
    full_name="fetch.aea.Fipa.FipaMessage.Inform_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="info",
            full_name="fetch.aea.Fipa.FipaMessage.Inform_Performative.info",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[_FIPAMESSAGE_INFORM_PERFORMATIVE_INFOENTRY,],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1404,
    serialized_end=1543,
)

_FIPAMESSAGE_ACCEPT_PERFORMATIVE = _descriptor.Descriptor(
    name="Accept_Performative",
    full_name="fetch.aea.Fipa.FipaMessage.Accept_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1545,
    serialized_end=1566,
)

_FIPAMESSAGE_DECLINE_PERFORMATIVE = _descriptor.Descriptor(
    name="Decline_Performative",
    full_name="fetch.aea.Fipa.FipaMessage.Decline_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1568,
    serialized_end=1590,
)

_FIPAMESSAGE_MATCH_ACCEPT_PERFORMATIVE = _descriptor.Descriptor(
    name="Match_Accept_Performative",
    full_name="fetch.aea.Fipa.FipaMessage.Match_Accept_Performative",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1592,
    serialized_end=1619,
)

_FIPAMESSAGE = _descriptor.Descriptor(
    name="FipaMessage",
    full_name="fetch.aea.Fipa.FipaMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="message_id",
            full_name="fetch.aea.Fipa.FipaMessage.message_id",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="dialogue_starter_reference",
            full_name="fetch.aea.Fipa.FipaMessage.dialogue_starter_reference",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="dialogue_responder_reference",
            full_name="fetch.aea.Fipa.FipaMessage.dialogue_responder_reference",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="target",
            full_name="fetch.aea.Fipa.FipaMessage.target",
            index=3,
            number=4,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="accept",
            full_name="fetch.aea.Fipa.FipaMessage.accept",
            index=4,
            number=5,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="accept_w_inform",
            full_name="fetch.aea.Fipa.FipaMessage.accept_w_inform",
            index=5,
            number=6,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="cfp",
            full_name="fetch.aea.Fipa.FipaMessage.cfp",
            index=6,
            number=7,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="decline",
            full_name="fetch.aea.Fipa.FipaMessage.decline",
            index=7,
            number=8,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="inform",
            full_name="fetch.aea.Fipa.FipaMessage.inform",
            index=8,
            number=9,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="match_accept",
            full_name="fetch.aea.Fipa.FipaMessage.match_accept",
            index=9,
            number=10,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="match_accept_w_inform",
            full_name="fetch.aea.Fipa.FipaMessage.match_accept_w_inform",
            index=10,
            number=11,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="propose",
            full_name="fetch.aea.Fipa.FipaMessage.propose",
            index=11,
            number=12,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[
        _FIPAMESSAGE_DESCRIPTION,
        _FIPAMESSAGE_QUERY,
        _FIPAMESSAGE_CFP_PERFORMATIVE,
        _FIPAMESSAGE_PROPOSE_PERFORMATIVE,
        _FIPAMESSAGE_ACCEPT_W_INFORM_PERFORMATIVE,
        _FIPAMESSAGE_MATCH_ACCEPT_W_INFORM_PERFORMATIVE,
        _FIPAMESSAGE_INFORM_PERFORMATIVE,
        _FIPAMESSAGE_ACCEPT_PERFORMATIVE,
        _FIPAMESSAGE_DECLINE_PERFORMATIVE,
        _FIPAMESSAGE_MATCH_ACCEPT_PERFORMATIVE,
    ],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="performative",
            full_name="fetch.aea.Fipa.FipaMessage.performative",
            index=0,
            containing_type=None,
            fields=[],
        ),
    ],
    serialized_start=31,
    serialized_end=1635,
)

_FIPAMESSAGE_DESCRIPTION.containing_type = _FIPAMESSAGE
_FIPAMESSAGE_QUERY_NOTHING.containing_type = _FIPAMESSAGE_QUERY
_FIPAMESSAGE_QUERY.fields_by_name["nothing"].message_type = _FIPAMESSAGE_QUERY_NOTHING
_FIPAMESSAGE_QUERY.containing_type = _FIPAMESSAGE
_FIPAMESSAGE_QUERY.oneofs_by_name["query"].fields.append(
    _FIPAMESSAGE_QUERY.fields_by_name["bytes"]
)
_FIPAMESSAGE_QUERY.fields_by_name[
    "bytes"
].containing_oneof = _FIPAMESSAGE_QUERY.oneofs_by_name["query"]
_FIPAMESSAGE_QUERY.oneofs_by_name["query"].fields.append(
    _FIPAMESSAGE_QUERY.fields_by_name["nothing"]
)
_FIPAMESSAGE_QUERY.fields_by_name[
    "nothing"
].containing_oneof = _FIPAMESSAGE_QUERY.oneofs_by_name["query"]
_FIPAMESSAGE_QUERY.oneofs_by_name["query"].fields.append(
    _FIPAMESSAGE_QUERY.fields_by_name["query_bytes"]
)
_FIPAMESSAGE_QUERY.fields_by_name[
    "query_bytes"
].containing_oneof = _FIPAMESSAGE_QUERY.oneofs_by_name["query"]
_FIPAMESSAGE_CFP_PERFORMATIVE.fields_by_name["query"].message_type = _FIPAMESSAGE_QUERY
_FIPAMESSAGE_CFP_PERFORMATIVE.containing_type = _FIPAMESSAGE
_FIPAMESSAGE_PROPOSE_PERFORMATIVE.fields_by_name[
    "proposal"
].message_type = _FIPAMESSAGE_DESCRIPTION
_FIPAMESSAGE_PROPOSE_PERFORMATIVE.containing_type = _FIPAMESSAGE
_FIPAMESSAGE_ACCEPT_W_INFORM_PERFORMATIVE_INFOENTRY.containing_type = (
    _FIPAMESSAGE_ACCEPT_W_INFORM_PERFORMATIVE
)
_FIPAMESSAGE_ACCEPT_W_INFORM_PERFORMATIVE.fields_by_name[
    "info"
].message_type = _FIPAMESSAGE_ACCEPT_W_INFORM_PERFORMATIVE_INFOENTRY
_FIPAMESSAGE_ACCEPT_W_INFORM_PERFORMATIVE.containing_type = _FIPAMESSAGE
_FIPAMESSAGE_MATCH_ACCEPT_W_INFORM_PERFORMATIVE_INFOENTRY.containing_type = (
    _FIPAMESSAGE_MATCH_ACCEPT_W_INFORM_PERFORMATIVE
)
_FIPAMESSAGE_MATCH_ACCEPT_W_INFORM_PERFORMATIVE.fields_by_name[
    "info"
].message_type = _FIPAMESSAGE_MATCH_ACCEPT_W_INFORM_PERFORMATIVE_INFOENTRY
_FIPAMESSAGE_MATCH_ACCEPT_W_INFORM_PERFORMATIVE.containing_type = _FIPAMESSAGE
_FIPAMESSAGE_INFORM_PERFORMATIVE_INFOENTRY.containing_type = (
    _FIPAMESSAGE_INFORM_PERFORMATIVE
)
_FIPAMESSAGE_INFORM_PERFORMATIVE.fields_by_name[
    "info"
].message_type = _FIPAMESSAGE_INFORM_PERFORMATIVE_INFOENTRY
_FIPAMESSAGE_INFORM_PERFORMATIVE.containing_type = _FIPAMESSAGE
_FIPAMESSAGE_ACCEPT_PERFORMATIVE.containing_type = _FIPAMESSAGE
_FIPAMESSAGE_DECLINE_PERFORMATIVE.containing_type = _FIPAMESSAGE
_FIPAMESSAGE_MATCH_ACCEPT_PERFORMATIVE.containing_type = _FIPAMESSAGE
_FIPAMESSAGE.fields_by_name["accept"].message_type = _FIPAMESSAGE_ACCEPT_PERFORMATIVE
_FIPAMESSAGE.fields_by_name[
    "accept_w_inform"
].message_type = _FIPAMESSAGE_ACCEPT_W_INFORM_PERFORMATIVE
_FIPAMESSAGE.fields_by_name["cfp"].message_type = _FIPAMESSAGE_CFP_PERFORMATIVE
_FIPAMESSAGE.fields_by_name["decline"].message_type = _FIPAMESSAGE_DECLINE_PERFORMATIVE
_FIPAMESSAGE.fields_by_name["inform"].message_type = _FIPAMESSAGE_INFORM_PERFORMATIVE
_FIPAMESSAGE.fields_by_name[
    "match_accept"
].message_type = _FIPAMESSAGE_MATCH_ACCEPT_PERFORMATIVE
_FIPAMESSAGE.fields_by_name[
    "match_accept_w_inform"
].message_type = _FIPAMESSAGE_MATCH_ACCEPT_W_INFORM_PERFORMATIVE
_FIPAMESSAGE.fields_by_name["propose"].message_type = _FIPAMESSAGE_PROPOSE_PERFORMATIVE
_FIPAMESSAGE.oneofs_by_name["performative"].fields.append(
    _FIPAMESSAGE.fields_by_name["accept"]
)
_FIPAMESSAGE.fields_by_name["accept"].containing_oneof = _FIPAMESSAGE.oneofs_by_name[
    "performative"
]
_FIPAMESSAGE.oneofs_by_name["performative"].fields.append(
    _FIPAMESSAGE.fields_by_name["accept_w_inform"]
)
_FIPAMESSAGE.fields_by_name[
    "accept_w_inform"
].containing_oneof = _FIPAMESSAGE.oneofs_by_name["performative"]
_FIPAMESSAGE.oneofs_by_name["performative"].fields.append(
    _FIPAMESSAGE.fields_by_name["cfp"]
)
_FIPAMESSAGE.fields_by_name["cfp"].containing_oneof = _FIPAMESSAGE.oneofs_by_name[
    "performative"
]
_FIPAMESSAGE.oneofs_by_name["performative"].fields.append(
    _FIPAMESSAGE.fields_by_name["decline"]
)
_FIPAMESSAGE.fields_by_name["decline"].containing_oneof = _FIPAMESSAGE.oneofs_by_name[
    "performative"
]
_FIPAMESSAGE.oneofs_by_name["performative"].fields.append(
    _FIPAMESSAGE.fields_by_name["inform"]
)
_FIPAMESSAGE.fields_by_name["inform"].containing_oneof = _FIPAMESSAGE.oneofs_by_name[
    "performative"
]
_FIPAMESSAGE.oneofs_by_name["performative"].fields.append(
    _FIPAMESSAGE.fields_by_name["match_accept"]
)
_FIPAMESSAGE.fields_by_name[
    "match_accept"
].containing_oneof = _FIPAMESSAGE.oneofs_by_name["performative"]
_FIPAMESSAGE.oneofs_by_name["performative"].fields.append(
    _FIPAMESSAGE.fields_by_name["match_accept_w_inform"]
)
_FIPAMESSAGE.fields_by_name[
    "match_accept_w_inform"
].containing_oneof = _FIPAMESSAGE.oneofs_by_name["performative"]
_FIPAMESSAGE.oneofs_by_name["performative"].fields.append(
    _FIPAMESSAGE.fields_by_name["propose"]
)
_FIPAMESSAGE.fields_by_name["propose"].containing_oneof = _FIPAMESSAGE.oneofs_by_name[
    "performative"
]
DESCRIPTOR.message_types_by_name["FipaMessage"] = _FIPAMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

FipaMessage = _reflection.GeneratedProtocolMessageType(
    "FipaMessage",
    (_message.Message,),
    {
        "Description": _reflection.GeneratedProtocolMessageType(
            "Description",
            (_message.Message,),
            {
                "DESCRIPTOR": _FIPAMESSAGE_DESCRIPTION,
                "__module__": "fipa_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Fipa.FipaMessage.Description)
            },
        ),
        "Query": _reflection.GeneratedProtocolMessageType(
            "Query",
            (_message.Message,),
            {
                "Nothing": _reflection.GeneratedProtocolMessageType(
                    "Nothing",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _FIPAMESSAGE_QUERY_NOTHING,
                        "__module__": "fipa_pb2"
                        # @@protoc_insertion_point(class_scope:fetch.aea.Fipa.FipaMessage.Query.Nothing)
                    },
                ),
                "DESCRIPTOR": _FIPAMESSAGE_QUERY,
                "__module__": "fipa_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Fipa.FipaMessage.Query)
            },
        ),
        "Cfp_Performative": _reflection.GeneratedProtocolMessageType(
            "Cfp_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _FIPAMESSAGE_CFP_PERFORMATIVE,
                "__module__": "fipa_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Fipa.FipaMessage.Cfp_Performative)
            },
        ),
        "Propose_Performative": _reflection.GeneratedProtocolMessageType(
            "Propose_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _FIPAMESSAGE_PROPOSE_PERFORMATIVE,
                "__module__": "fipa_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Fipa.FipaMessage.Propose_Performative)
            },
        ),
        "Accept_W_Inform_Performative": _reflection.GeneratedProtocolMessageType(
            "Accept_W_Inform_Performative",
            (_message.Message,),
            {
                "InfoEntry": _reflection.GeneratedProtocolMessageType(
                    "InfoEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _FIPAMESSAGE_ACCEPT_W_INFORM_PERFORMATIVE_INFOENTRY,
                        "__module__": "fipa_pb2"
                        # @@protoc_insertion_point(class_scope:fetch.aea.Fipa.FipaMessage.Accept_W_Inform_Performative.InfoEntry)
                    },
                ),
                "DESCRIPTOR": _FIPAMESSAGE_ACCEPT_W_INFORM_PERFORMATIVE,
                "__module__": "fipa_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Fipa.FipaMessage.Accept_W_Inform_Performative)
            },
        ),
        "Match_Accept_W_Inform_Performative": _reflection.GeneratedProtocolMessageType(
            "Match_Accept_W_Inform_Performative",
            (_message.Message,),
            {
                "InfoEntry": _reflection.GeneratedProtocolMessageType(
                    "InfoEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _FIPAMESSAGE_MATCH_ACCEPT_W_INFORM_PERFORMATIVE_INFOENTRY,
                        "__module__": "fipa_pb2"
                        # @@protoc_insertion_point(class_scope:fetch.aea.Fipa.FipaMessage.Match_Accept_W_Inform_Performative.InfoEntry)
                    },
                ),
                "DESCRIPTOR": _FIPAMESSAGE_MATCH_ACCEPT_W_INFORM_PERFORMATIVE,
                "__module__": "fipa_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Fipa.FipaMessage.Match_Accept_W_Inform_Performative)
            },
        ),
        "Inform_Performative": _reflection.GeneratedProtocolMessageType(
            "Inform_Performative",
            (_message.Message,),
            {
                "InfoEntry": _reflection.GeneratedProtocolMessageType(
                    "InfoEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _FIPAMESSAGE_INFORM_PERFORMATIVE_INFOENTRY,
                        "__module__": "fipa_pb2"
                        # @@protoc_insertion_point(class_scope:fetch.aea.Fipa.FipaMessage.Inform_Performative.InfoEntry)
                    },
                ),
                "DESCRIPTOR": _FIPAMESSAGE_INFORM_PERFORMATIVE,
                "__module__": "fipa_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Fipa.FipaMessage.Inform_Performative)
            },
        ),
        "Accept_Performative": _reflection.GeneratedProtocolMessageType(
            "Accept_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _FIPAMESSAGE_ACCEPT_PERFORMATIVE,
                "__module__": "fipa_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Fipa.FipaMessage.Accept_Performative)
            },
        ),
        "Decline_Performative": _reflection.GeneratedProtocolMessageType(
            "Decline_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _FIPAMESSAGE_DECLINE_PERFORMATIVE,
                "__module__": "fipa_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Fipa.FipaMessage.Decline_Performative)
            },
        ),
        "Match_Accept_Performative": _reflection.GeneratedProtocolMessageType(
            "Match_Accept_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _FIPAMESSAGE_MATCH_ACCEPT_PERFORMATIVE,
                "__module__": "fipa_pb2"
                # @@protoc_insertion_point(class_scope:fetch.aea.Fipa.FipaMessage.Match_Accept_Performative)
            },
        ),
        "DESCRIPTOR": _FIPAMESSAGE,
        "__module__": "fipa_pb2"
        # @@protoc_insertion_point(class_scope:fetch.aea.Fipa.FipaMessage)
    },
)
_sym_db.RegisterMessage(FipaMessage)
_sym_db.RegisterMessage(FipaMessage.Description)
_sym_db.RegisterMessage(FipaMessage.Query)
_sym_db.RegisterMessage(FipaMessage.Query.Nothing)
_sym_db.RegisterMessage(FipaMessage.Cfp_Performative)
_sym_db.RegisterMessage(FipaMessage.Propose_Performative)
_sym_db.RegisterMessage(FipaMessage.Accept_W_Inform_Performative)
_sym_db.RegisterMessage(FipaMessage.Accept_W_Inform_Performative.InfoEntry)
_sym_db.RegisterMessage(FipaMessage.Match_Accept_W_Inform_Performative)
_sym_db.RegisterMessage(FipaMessage.Match_Accept_W_Inform_Performative.InfoEntry)
_sym_db.RegisterMessage(FipaMessage.Inform_Performative)
_sym_db.RegisterMessage(FipaMessage.Inform_Performative.InfoEntry)
_sym_db.RegisterMessage(FipaMessage.Accept_Performative)
_sym_db.RegisterMessage(FipaMessage.Decline_Performative)
_sym_db.RegisterMessage(FipaMessage.Match_Accept_Performative)


_FIPAMESSAGE_ACCEPT_W_INFORM_PERFORMATIVE_INFOENTRY._options = None
_FIPAMESSAGE_MATCH_ACCEPT_W_INFORM_PERFORMATIVE_INFOENTRY._options = None
_FIPAMESSAGE_INFORM_PERFORMATIVE_INFOENTRY._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains a test for aea.test_tools.test_cases."""

from pathlib import Path
from typing import cast

import pytest

from aea.exceptions import AEAEnforceError
from aea.mail.base import Address
from aea.protocols.base import Message
from aea.protocols.default.message import DefaultMessage
from aea.protocols.dialogue.base import Dialogue, DialogueLabel, DialogueMessage
from aea.skills.base import Skill
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.protocols.fipa.dialogues import FipaDialogue
from packages.fetchai.protocols.fipa.dialogues import FipaDialogues as BaseFipaDialogues
from packages.fetchai.protocols.fipa.message import FipaMessage

from tests.conftest import ROOT_DIR


class TestSkillTestCase(BaseSkillTestCase):
    """Test case for BaseSkillTestCase."""

    path_to_skill = Path(ROOT_DIR, "tests", "data", "dummy_skill")

    def test_setup(self):
        """Test the setup() class method."""
        assert self.skill.skill_context.agent_address == "test_agent_address"
        assert self.skill.skill_context.agent_name == "test_agent_name"
        assert (
            self.skill.skill_context.search_service_address
            == "dummy_search_service_address"
        )
        assert (
            self.skill.skill_context.decision_maker_address
            == "dummy_decision_maker_address"
        )
        assert "dummy" in self.skill.behaviours.keys()
        assert "dummy" in self.skill.handlers.keys()
        assert "dummy_internal" in self.skill.handlers.keys()
        assert "dummy" in self.skill.models.keys()

    def test_properties(self):
        """Test the properties."""
        assert isinstance(self.skill, Skill)
        assert self.skill.behaviours.get("dummy") is not None

    def test_get_quantity_in_outbox(self):
        """Test the get_quantity_in_outbox method."""
        assert self.get_quantity_in_outbox() == 0

        dummy_message = Message(dummy="dummy")
        dummy_message.to = "some_to"
        dummy_message.sender = "some_sender"
        self.skill.skill_context.outbox.put_message(dummy_message)

        assert self.get_quantity_in_outbox() == 1

    def test_get_message_from_outbox(self):
        """Test the get_message_from_outbox method."""
        assert self.get_message_from_outbox() is None

        dummy_message_1 = Message(dummy_1="dummy_1")
        dummy_message_1.to = "some_to_1"
        dummy_message_1.sender = "some_sender_1"
        self.skill.skill_context.outbox.put_message(dummy_message_1)

        dummy_message_2 = Message(dummy_2="dummy_2")
        dummy_message_2.to = "some_to_2"
        dummy_message_2.sender = "some_sender_2"
        self.skill.skill_context.outbox.put_message(dummy_message_2)

        assert self.get_message_from_outbox() == dummy_message_1
        assert self.get_message_from_outbox() == dummy_message_2

    def test_get_quantity_in_decision_maker_inbox(self):
        """Test the get_quantity_in_decision_maker_inbox method."""
        assert self.get_quantity_in_decision_maker_inbox() == 0

        dummy_message = Message(dummy="dummy")
        dummy_message.to = "some_to"
        dummy_message.sender = "some_sender"
        self.skill.skill_context.decision_maker_message_queue.put(dummy_message)

        assert self.get_quantity_in_decision_maker_inbox() == 1

    def test_get_message_from_decision_maker_inbox(self):
        """Test the get_message_from_decision_maker_inbox method."""
        assert self.get_message_from_decision_maker_inbox() is None

        dummy_message_1 = Message(dummy_1="dummy_1")
        dummy_message_1.to = "some_to_1"
        dummy_message_1.sender = "some_sender_1"
        self.skill.skill_context.decision_maker_message_queue.put(dummy_message_1)

        dummy_message_2 = Message(dummy_2="dummy_2")
        dummy_message_2.to = "some_to_2"
        dummy_message_2.sender = "some_sender_2"
        self.skill.skill_context.decision_maker_message_queue.put(dummy_message_2)

        assert self.get_message_from_decision_maker_inbox() == dummy_message_1
        assert self.get_message_from_decision_maker_inbox() == dummy_message_2

    def test_assert_quantity_in_outbox(self):
        """Test the assert_quantity_in_outbox method."""
        with pytest.raises(
            AssertionError,
            match=f"Invalid number of messages in outbox. Expected {1}. Found {0}.",
        ):
            self.assert_quantity_in_outbox(1)

        dummy_message = Message(dummy="dummy")
        dummy_message.to = "some_to"
        dummy_message.sender = "some_sender"
        self.skill.skill_context.outbox.put_message(dummy_message)

        self.assert_quantity_in_outbox(1)

    def test_assert_quantity_in_decision_making_queue(self):
        """Test the assert_quantity_in_decision_making_queue method."""
        with pytest.raises(
            AssertionError,
            match=f"Invalid number of messages in decision maker queue. Expected {1}. Found {0}.",
        ):
            self.assert_quantity_in_decision_making_queue(1)

        dummy_message = Message(dummy_1="dummy_1")
        dummy_message.to = "some_to_1"
        dummy_message.sender = "some_sender_1"
        self.skill.skill_context.decision_maker_message_queue.put(dummy_message)

        self.assert_quantity_in_decision_making_queue(1)

    def test_positive_message_has_attributes_valid_type(self):
        """Test the message_has_attributes method where the message is of the specified type."""
        dummy_message = FipaMessage(
            dialogue_reference=("0", "0"),
            message_id=1,
            performative=FipaMessage.Performative.CFP,
            target=0,
            query="some_query",
        )
        dummy_message.to = "some_to"
        dummy_message.sender = "some_sender"

        valid_has_attribute, valid_has_attribute_msg = self.message_has_attributes(
            actual_message=dummy_message,
            message_type=FipaMessage,
            message_id=1,
            performative=FipaMessage.Performative.CFP,
            target=0,
            query="some_query",
            to="some_to",
            sender="some_sender",
        )
        assert valid_has_attribute
        assert (
            valid_has_attribute_msg
            == "The message has the provided expected attributes."
        )

    def test_negative_message_has_attributes_invalid_message_id(self):
        """Negative test for message_has_attributes method where the message id does NOT match."""
        dummy_message = FipaMessage(
            dialogue_reference=("0", "0"),
            message_id=1,
            performative=FipaMessage.Performative.CFP,
            target=0,
            query="some_query",
        )
        dummy_message.to = "some_to"
        dummy_message.sender = "some_sender"

        invalid_has_attribute, invalid_has_attribute_msg = self.message_has_attributes(
            actual_message=dummy_message,
            message_type=FipaMessage,
            message_id=2,
            performative=FipaMessage.Performative.CFP,
            target=0,
            query="some_query",
            to="some_to",
            sender="some_sender",
        )
        assert not invalid_has_attribute
        assert (
            invalid_has_attribute_msg
            == "The 'message_id' fields do not match. Actual 'message_id': 1. Expected 'message_id': 2"
        )

    def test_negative_message_has_attributes_invalid_type(self):
        """Test the message_has_attributes method where the message is NOT of the specified type."""
        dummy_message = FipaMessage(
            dialogue_reference=("0", "0"),
            message_id=1,
            performative=FipaMessage.Performative.CFP,
            target=0,
            query="some_query",
        )
        dummy_message.to = "some_to"
        dummy_message.sender = "some_sender"

        valid_has_attribute, valid_has_attribute_msg = self.message_has_attributes(
            actual_message=dummy_message,
            message_type=DefaultMessage,
            message_id=1,
            performative=FipaMessage.Performative.CFP,
            target=0,
            query="some_query",
            to="some_to",
            sender="some_sender",
        )
        assert not valid_has_attribute
        assert (
            valid_has_attribute_msg
            == f"The message types do not match. Actual type: {FipaMessage}. Expected type: {DefaultMessage}"
        )

        invalid_has_attribute, invalid_has_attribute_msg = self.message_has_attributes(
            actual_message=dummy_message,
            message_type=FipaMessage,
            message_id=2,
            performative=FipaMessage.Performative.CFP,
            target=0,
            query="some_query",
            to="some_to",
            sender="some_sender",
        )
        assert not invalid_has_attribute
        assert (
            invalid_has_attribute_msg
            == "The 'message_id' fields do not match. Actual 'message_id': 1. Expected 'message_id': 2"
        )

    def test_build_incoming_message(self):
        """Test the build_incoming_message method."""
        message_type = FipaMessage
        performative = FipaMessage.Performative.CFP
        dialogue_reference = ("1", "1")
        to = "some_to"
        query = "some_query"
        incoming_message = self.build_incoming_message(
            message_type=message_type,
            performative=performative,
            dialogue_reference=dialogue_reference,
            to=to,
            query=query,
        )

        assert type(incoming_message) == message_type
        incoming_message = cast(FipaMessage, incoming_message)
        assert incoming_message.dialogue_reference == dialogue_reference
        assert incoming_message.message_id == 1
        assert incoming_message.target == 0
        assert incoming_message.performative == performative
        assert incoming_message.query == query
        assert incoming_message.sender == "counterparty"
        assert incoming_message.to == to

    def test_positive_build_incoming_message_for_skill_dialogue(self):
        """Positive test for build_incoming_message_for_skill_dialogue method."""
        fipa_dialogues = FipaDialogues(
            self_address=self.skill.skill_context.agent_address
        )
        _, dialogue = fipa_dialogues.create(
            counterparty="some_counterparty",
            performative=FipaMessage.Performative.CFP,
            query="some_query",
        )

        performative = FipaMessage.Performative.PROPOSE
        proposal = "some_proposal"
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=dialogue, performative=performative, proposal=proposal,
        )

        assert type(incoming_message) == FipaMessage
        incoming_message = cast(FipaMessage, incoming_message)
        assert (
            incoming_message.dialogue_reference
            == dialogue.dialogue_label.dialogue_reference
        )
        assert incoming_message.message_id == 2
        assert incoming_message.target == 1
        assert incoming_message.performative == performative
        assert incoming_message.proposal == proposal
        assert incoming_message.sender == dialogue.dialogue_label.dialogue_opponent_addr
        assert incoming_message.to == dialogue.self_address

    def test_negative_build_incoming_message_for_skill_dialogue_dialogue_is_none(self):
        """Negative test for build_incoming_message_for_skill_dialogue method where the provided dialogue is None."""
        performative = FipaMessage.Performative.PROPOSE
        proposal = "some_proposal"

        with pytest.raises(AEAEnforceError, match="dialogue cannot be None."):
            self.build_incoming_message_for_skill_dialogue(
                dialogue=None, performative=performative, proposal=proposal,
            )

    def test_negative_build_incoming_message_for_skill_dialogue_dialogue_is_empty(self):
        """Negative test for build_incoming_message_for_skill_dialogue method where the provided dialogue is empty."""
        performative = FipaMessage.Performative.PROPOSE
        proposal = "some_proposal"

        fipa_dialogues = FipaDialogues(
            self_address=self.skill.skill_context.agent_address
        )
        dialogue = fipa_dialogues._create_self_initiated(
            dialogue_opponent_addr="some_counterparty",
            dialogue_reference=("0", ""),
            role=FipaDialogue.Role.BUYER,
        )

        with pytest.raises(AEAEnforceError, match="dialogue cannot be empty."):
            self.build_incoming_message_for_skill_dialogue(
                dialogue=dialogue, performative=performative, proposal=proposal,
            )

    def test_provide_unspecified_fields(self):
        """Test the _provide_unspecified_fields method."""
        dialogue_message_unspecified = DialogueMessage(
            FipaMessage.Performative.ACCEPT, {}
        )

        is_incoming, target = self._provide_unspecified_fields(
            dialogue_message_unspecified, last_is_incoming=False, message_id=2
        )
        assert is_incoming is True
        assert target == 1

        dialogue_message_specified = DialogueMessage(
            FipaMessage.Performative.ACCEPT, {}, False, 4
        )

        is_incoming, target = self._provide_unspecified_fields(
            dialogue_message_specified, last_is_incoming=True, message_id=7
        )
        assert is_incoming is False
        assert target == 4

    def test_non_initial_incoming_message_dialogue_reference(self):
        """Test the _non_initial_incoming_message_dialogue_reference method."""
        dialogue_incomplete_ref = FipaDialogue(
            DialogueLabel(("2", ""), "opponent", "self_address"),
            "self_address",
            FipaDialogue.Role.BUYER,
        )
        reference_incomplete = self._non_initial_incoming_message_dialogue_reference(
            dialogue_incomplete_ref
        )
        assert reference_incomplete[1] != ""

        dialogue_complete_ref = FipaDialogue(
            DialogueLabel(("2", "7"), "opponent", "self_address"),
            "self_address",
            FipaDialogue.Role.BUYER,
        )
        reference_complete = self._non_initial_incoming_message_dialogue_reference(
            dialogue_complete_ref
        )
        assert reference_complete[1] == "7"

    def test_extract_message_fields(self):
        """Test the _extract_message_fields method."""
        expected_performative = FipaMessage.Performative.ACCEPT
        expected_contents = {}
        expected_is_incoming = False
        expected_target = 4
        dialogue_message = DialogueMessage(
            expected_performative,
            expected_contents,
            expected_is_incoming,
            expected_target,
        )

        (
            actual_performative,
            actual_contents,
            actual_message_id,
            actual_is_incoming,
            actual_target,
        ) = self._extract_message_fields(
            message=dialogue_message, index=3, last_is_incoming=True
        )

        assert actual_message_id == 4
        assert actual_target == expected_target
        assert actual_performative == expected_performative
        assert actual_contents == expected_contents
        assert actual_is_incoming == expected_is_incoming

    def test_prepare_skill_dialogue_valid_self_initiated(self):
        """Positive test for prepare_skill_dialogue method with a valid dialogue initiated by self."""
        fipa_dialogues = FipaDialogues(
            self_address=self.skill.skill_context.agent_address
        )
        dialogue_messages = (
            DialogueMessage(FipaMessage.Performative.CFP, {"query": "some_query"}),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE, {"proposal": "some_proposal"}
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_1"},
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_2"},
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_3"},
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_4"},
            ),
            DialogueMessage(FipaMessage.Performative.ACCEPT, {}),
            DialogueMessage(
                FipaMessage.Performative.MATCH_ACCEPT_W_INFORM, {"info": "some_info"}
            ),
        )
        dialogue = self.prepare_skill_dialogue(
            fipa_dialogues, dialogue_messages, "counterparty",
        )

        assert type(dialogue) == FipaDialogue
        assert dialogue.is_self_initiated
        assert len(dialogue._outgoing_messages) == 4
        assert len(dialogue._incoming_messages) == 4
        assert dialogue._get_message(4).proposal == "some_counter_proposal_2"
        assert dialogue._get_message(8).info == "some_info"

    def test_prepare_skill_dialogue_valid_opponent_initiated(self):
        """Positive test for prepare_skill_dialogue method with a valid dialogue initiated by the opponent."""
        fipa_dialogues = FipaDialogues(
            self_address=self.skill.skill_context.agent_address
        )
        dialogue_messages = (
            DialogueMessage(
                FipaMessage.Performative.CFP, {"query": "some_query"}, True
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE, {"proposal": "some_proposal"}
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_1"},
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_2"},
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_3"},
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_4"},
            ),
            DialogueMessage(FipaMessage.Performative.ACCEPT, {}),
            DialogueMessage(
                FipaMessage.Performative.MATCH_ACCEPT_W_INFORM, {"info": "some_info"}
            ),
        )
        dialogue = self.prepare_skill_dialogue(
            fipa_dialogues, dialogue_messages, "counterparty",
        )

        assert type(dialogue) == FipaDialogue
        assert not dialogue.is_self_initiated
        assert len(dialogue._outgoing_messages) == 4
        assert len(dialogue._incoming_messages) == 4
        assert dialogue._get_message(4).proposal == "some_counter_proposal_2"
        assert dialogue._get_message(8).info == "some_info"

    def test_negative_prepare_skill_dialogue_invalid_opponent_initiated(self):
        """Negative test for prepare_skill_dialogue method with an invalid dialogue initiated by the opponent."""
        fipa_dialogues = FipaDialogues(
            self_address=self.skill.skill_context.agent_address
        )
        dialogue_messages = (
            DialogueMessage(
                FipaMessage.Performative.PROPOSE, {"proposal": "some_proposal"}, True
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_1"},
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_2"},
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_3"},
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_4"},
            ),
            DialogueMessage(FipaMessage.Performative.ACCEPT, {}),
            DialogueMessage(
                FipaMessage.Performative.MATCH_ACCEPT_W_INFORM, {"info": "some_info"}
            ),
        )
        with pytest.raises(
            AEAEnforceError, match="Cannot update the dialogue with message number 1"
        ):
            self.prepare_skill_dialogue(
                fipa_dialogues, dialogue_messages, "counterparty",
            )

    def test_negative_prepare_skill_dialogue_empty_messages(self):
        """Negative test for prepare_skill_dialogue method where the list of DialogueMessages is emoty."""
        fipa_dialogues = FipaDialogues(
            self_address=self.skill.skill_context.agent_address
        )
        dialogue_messages = tuple()

        with pytest.raises(
            AEAEnforceError, match="the list of messages must be positive."
        ):
            self.prepare_skill_dialogue(
                fipa_dialogues, dialogue_messages, "counterparty",
            )

    def test_negative_prepare_skill_dialogue_invalid(self):
        """Negative test for prepare_skill_dialogue method with an invalid dialogue (a message has invalid target)."""
        fipa_dialogues = FipaDialogues(
            self_address=self.skill.skill_context.agent_address
        )
        dialogue_messages = (
            DialogueMessage(FipaMessage.Performative.CFP, {"query": "some_query"}),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_proposal"},
                target=2,
            ),
        )

        with pytest.raises(
            AEAEnforceError, match="Cannot update the dialogue with message number 2"
        ):
            self.prepare_skill_dialogue(
                fipa_dialogues, dialogue_messages, "counterparty",
            )


class FipaDialogues(BaseFipaDialogues):
    """The dialogues class keeps track of all dialogues."""

    def __init__(self, self_address: Address, **kwargs) -> None:
        """
        Initialize dialogues.

        :return: None
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> Dialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return FipaDialogue.Role.BUYER

        BaseFipaDialogues.__init__(
            self,
            self_address=self_address,
            role_from_first_message=role_from_first_message,
            dialogue_class=FipaDialogue,
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Methods for CLI push functionality."""

import os
import shutil
import tarfile

import click

from aea.cli.registry.utils import (
    check_is_author_logged_in,
    clean_tarfiles,
    request_api,
)
from aea.cli.utils.context import Context
from aea.cli.utils.generic import is_readme_present, load_yaml
from aea.cli.utils.loggers import logger
from aea.configurations.base import DEFAULT_README_FILE, PublicId


def _remove_pycache(source_dir: str):
    pycache_path = os.path.join(source_dir, "__pycache__")
    if os.path.exists(pycache_path):
        shutil.rmtree(pycache_path)


def _compress_dir(output_filename: str, source_dir: str):
    _remove_pycache(source_dir)
    with tarfile.open(output_filename, "w:gz") as f:
        f.add(source_dir, arcname=os.path.basename(source_dir))


@clean_tarfiles
def push_item(ctx: Context, item_type: str, item_id: PublicId) -> None:
    """
    Push item to the Registry.

    :param item_type: str type of item (connection/protocol/skill).
    :param item_id: str item name.

    :return: None
    """
    item_type_plural = item_type + "s"

    items_folder = os.path.join(ctx.cwd, item_type_plural)
    item_path = os.path.join(items_folder, item_id.name)

    item_config_filepath = os.path.join(item_path, "{}.yaml".format(item_type))
    logger.debug("Reading {} {} config ...".format(item_id.name, item_type))
    item_config = load_yaml(item_config_filepath)
    check_is_author_logged_in(item_config["author"])

    logger.debug(
        "Searching for {} {} in {} ...".format(item_id.name, item_type, items_folder)
    )
    if not os.path.exists(item_path):
        raise click.ClickException(
            '{} "{}" not found  in {}. Make sure you run push command '
            "from a correct folder.".format(
                item_type.title(), item_id.name, items_folder
            )
        )

    output_filename = "{}.tar.gz".format(item_id.name)
    logger.debug(
        "Compressing {} {} to {} ...".format(item_id.name, item_type, output_filename)
    )
    _compress_dir(output_filename, item_path)
    output_filepath = os.path.join(ctx.cwd, output_filename)

    data = {
        "name": item_id.name,
        "description": item_config["description"],
        "version": item_config["version"],
    }

    # dependencies
    for key in ["connections", "contracts", "protocols", "skills"]:
        deps_list = item_config.get(key)
        if deps_list:
            data.update({key: deps_list})

    files = {"file": open(output_filepath, "rb")}
    readme_path = os.path.join(item_path, DEFAULT_README_FILE)
    if is_readme_present(readme_path):
        files["readme"] = open(readme_path, "rb")

    path = "/{}/create".format(item_type_plural)
    logger.debug("Pushing {} {} to Registry ...".format(item_id.name, item_type))
    resp = request_api("POST", path, data=data, is_auth=True, files=files)
    click.echo(
        "Successfully pushed {} {} to the Registry. Public ID: {}".format(
            item_type, item_id.name, resp["public_id"]
        )
    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains dummy weather station data."""

import datetime
import logging
import os.path
import random
import sqlite3
import time
from typing import Dict, Union


_default_logger = logging.getLogger(
    "aea.packages.fetchai.skills.weather_station.dummy_weather_station_data"
)

my_path = os.path.dirname(__file__)

DB_SOURCE = os.path.join(my_path, "dummy_weather_station_data.db")

# Checking if the database exists
con = sqlite3.connect(DB_SOURCE)
cur = con.cursor()

cur.close()
con.commit()
con.close()

# Create a table if it doesn't exist'
command = """ CREATE TABLE IF NOT EXISTS data (
                                 abs_pressure REAL,
                                 delay REAL,
                                 hum_in REAL,
                                 hum_out REAL,
                                 idx TEXT,
                                 rain REAL,
                                 temp_in REAL,
                                 temp_out REAL,
                                 wind_ave REAL,
                                 wind_dir REAL,
                                 wind_gust REAL)"""

con = sqlite3.connect(DB_SOURCE)
cur = con.cursor()
cur.execute(command)
cur.close()
con.commit()
if con is not None:
    _default_logger.debug(
        "Weather station: I closed the db after checking it is populated!"
    )
    con.close()


class Forecast:
    """Represents a whether forecast."""

    @staticmethod
    def add_data(tagged_data: Dict[str, Union[int, datetime.datetime]]) -> None:
        """
        Add data to the forecast.

        :param tagged_data: the data dictionary
        :return: None
        """
        con_ = sqlite3.connect(DB_SOURCE)
        cur_ = con_.cursor()
        cur_.execute(
            """INSERT INTO data(abs_pressure,
                                       delay,
                                       hum_in,
                                       hum_out,
                                       idx,
                                       rain,
                                       temp_in,
                                       temp_out,
                                       wind_ave,
                                       wind_dir,
                                       wind_gust) VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)""",
            (
                tagged_data["abs_pressure"],
                tagged_data["delay"],
                tagged_data["hum_in"],
                tagged_data["hum_out"],
                datetime.datetime.now().strftime("%s"),
                tagged_data["rain"],
                tagged_data["temp_in"],
                tagged_data["temp_out"],
                tagged_data["wind_ave"],
                tagged_data["wind_dir"],
                tagged_data["wind_gust"],
            ),
        )
        _default_logger.info("Wheather station: I added data in the db!")
        cur_.close()
        con_.commit()
        con_.close()

    def generate(self):
        """Generate weather data."""
        while True:  # nosec
            dict_of_data = {
                "abs_pressure": random.randrange(1022.0, 1025, 1),
                "delay": random.randint(2, 7),
                "hum_in": random.randrange(33.0, 40.0, 1),
                "hum_out": random.randrange(33.0, 80.0, 1),
                "idx": datetime.datetime.now(),
                "rain": random.randrange(70.0, 74.0, 1),
                "temp_in": random.randrange(18, 28, 1),
                "temp_out": random.randrange(2, 20, 1),
                "wind_ave": random.randrange(0, 10, 1),
                "wind_dir": random.randrange(0, 14, 1),
                "wind_gust": random.randrange(1, 7, 1),
            }  # type: Dict[str, Union[int, datetime.datetime]]
            self.add_data(dict_of_data)
            time.sleep(5)


if __name__ == "__main__":
    a = Forecast()
    a.generate()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains tests for aea.crypto.registries"""

from typing import Optional

from aea.crypto.base import Crypto
from aea.crypto.cosmos import CosmosApi, CosmosCrypto
from aea.crypto.registries import make_crypto, make_ledger_api
from aea.crypto.registries.base import ItemId, Registry


COSMOS = CosmosCrypto.identifier


def test_make_crypto_cosmos_positive():
    """Test make_crypto for fetchai."""
    crypto = make_crypto(COSMOS)
    assert isinstance(crypto, CosmosCrypto)


def test_make_ledger_api_cosmos_positive():
    """Test make_crypto for fetchai."""
    ledger_api = make_ledger_api(COSMOS, **{"network": "testnet"})
    assert isinstance(ledger_api, CosmosApi)


class Something:
    """Some class."""

    class_key = None  # type: Optional[str]

    def __init__(self, **kwargs):
        """Initialize something."""
        self.kwargs = kwargs


def test_register_make_with_class_kwargs():
    """Test registry make with class kwargs."""
    reg = Registry()
    id_ = "id"
    kwargs = {"key": "value"}
    class_kwargs = {"class_key": "class_value"}
    reg.register(
        id_=id_,
        entry_point="tests.test_crypto.test_registries:Something",
        class_kwargs=class_kwargs,
        **kwargs
    )
    assert Something.class_key is None
    item = reg.make(id_)
    assert item is not None
    assert type(item) == Something
    assert item.kwargs == kwargs
    assert item.class_key == "class_value"


def test_itemid():
    """Test the idemid object."""
    item_id = ItemId(COSMOS)
    assert item_id.name == COSMOS


def test_registry():
    """Test the registry object."""
    registry = Registry[Crypto]()
    item_id = ItemId(COSMOS)
    assert not registry.has_spec(item_id), "Registry should be empty"
